import zebrazoom.code.paths as paths
from zebrazoom.dataAnalysis.datasetcreation.createDataFrame import createDataFrame
import pandas as pd
import os

def createPandasDataFrameOfParameters(hyperparameters, videoName, videoExtension, ZZoutputLocation='', supstructOverwrite={}):
  
  print("Creating pandas dataframe of parameters")
  
  nbWells = hyperparameters["nbWells"]
    
  excelFileDataFrame = pd.DataFrame(data=[["defaultZZoutputFolder", videoName, hyperparameters["videoFPS"], hyperparameters["videoPixelSize"], str([1 for i in range(nbWells)]), str([1 for i in range(nbWells)]), str([1 for i in range(nbWells)])]], columns=['path', 'trial_id', 'fq', 'pixelsize', 'condition', 'genotype', 'include'])
  
  if hyperparameters["frameStepForDistanceCalculation"] and hyperparameters["frameStepForDistanceCalculation"] != -1:
    frameStepForDistanceCalculation = int(hyperparameters["frameStepForDistanceCalculation"])
  else:
    frameStepForDistanceCalculation = 4 #int(hyperparameters["videoFPS"] / 100) * 4 if int(hyperparameters["videoFPS"]) >= 100 else 2
  
  print("Parameter: frameStepForDistanceCalculation:", frameStepForDistanceCalculation)
  
  if len(ZZoutputLocation) == 0:
    cur_dir_path = paths.getDefaultZZoutputFolder()
  else:
    cur_dir_path = ZZoutputLocation

  dataframeOptions = {
    'pathToExcelFile'                   : "",
    'fileExtension'                     : videoExtension,
    'resFolder'                         : os.path.join(cur_dir_path, videoName),
    'nameOfFile'                        : videoName,
    'smoothingFactorDynaParam'          : 0,
    'nbFramesTakenIntoAccount'          : -2, ## nbFramesTakenIntoAccount will be 100 minimum
    'numberOfBendsIncludedForMaxDetect' : -1,
    'minNbBendForBoutDetect'            : -1,
    'keepSpeedDistDurWhenLowNbBends'    : 1,
    'defaultZZoutputFolderPath'         : cur_dir_path,
    'getTailAngleSignMultNormalized'    : 1,
    'computeTailAngleParamForCluster'   : True,
    'computeMassCenterParamForCluster'  : True,
    'frameStepForDistanceCalculation'   : str(frameStepForDistanceCalculation),
    'tailAngleKinematicParameterCalculation'    : 1,
    'saveRawDataInAllBoutsSuperStructure'       : 1,
    'saveAllBoutsSuperStructuresInMatlabFormat' : 0
  }
  
  # forcePandasDfRecreation (third parameter) is set to True here
  [conditions, genotypes, nbFramesTakenIntoAccount, globParam] = createDataFrame(dataframeOptions, excelFileDataFrame, 1, [], 0, supstructOverwrite, H5filename=hyperparameters.get('H5filename', None))
  


import contextlib
import os
import shutil
import sys
import traceback
import tempfile
from datetime import datetime

import json

import numpy as np

from PyQt5.QtCore import Qt, QSize
from PyQt5.QtGui import QFont, QIcon
from PyQt5.QtWidgets import QWidget, QButtonGroup, QCheckBox, QDialog, QFileDialog, QHBoxLayout, QApplication, QLabel, QMainWindow, QRadioButton, QStackedLayout, QVBoxLayout, QMessageBox, QTextEdit, QSpacerItem, QPushButton
QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)
QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)
QApplication.setHighDpiScaleFactorRoundingPolicy(Qt.HighDpiScaleFactorRoundingPolicy.PassThrough)
PYQT6 = False

import zebrazoom.videoFormatConversion.zzVideoReading as zzVideoReading
import zebrazoom.code.paths as paths
import zebrazoom.code.util as util
import zebrazoom.code.GUI.configFilePrepareFunctions as configFilePrepareFunctions
import zebrazoom.code.GUI.GUI_InitialFunctions as GUI_InitialFunctions
import zebrazoom.code.GUI.configFileZebrafishFunctions as configFileZebrafishFunctions
import zebrazoom.code.GUI.adjustParameterInsideAlgoFunctions as adjustParameterInsideAlgoFunctions
import zebrazoom.code.GUI.dataAnalysisGUIFunctions as dataAnalysisGUIFunctions
import zebrazoom.code.GUI.troubleshootingFunction as troubleshootingFunction
from zebrazoom.zebraZoomVideoAnalysis import ZebraZoomVideoAnalysis
from zebrazoom.code.adjustHyperparameters import adjustFreelySwimTrackingParams, adjustFreelySwimTrackingAutoParams, adjustHeadEmbededTrackingParams, adjustHeadEmbeddedEyeTrackingParamsSegment, adjustHeadEmbeddedEyeTrackingParamsEllipse
from zebrazoom.code.GUI.GUI_InitialClasses import StartPage, VideoToAnalyze, ConfigFilePromp, Patience, ZZoutro, ZZoutroSbatch, SeveralVideos, FolderToAnalyze, TailExtremityHE, FolderMultipleROIInitialSelect, EnhanceZZOutput, ViewParameters, Error
from zebrazoom.code.GUI.configFilePrepare import ChooseVideoToCreateConfigFileFor, OptimizeConfigFile, ChooseGeneralExperiment, ChooseCenterOfMassTracking, WellOrganisation, FreelySwimmingExperiment, NbRegionsOfInterest, HomegeneousWellsLayout, CircularOrRectangularWells, NumberOfAnimals, NumberOfAnimals2, NumberOfAnimalsCenterOfMass, IdentifyHeadCenter, IdentifyBodyExtremity, FinishConfig, ChooseCircularWellsLeft, ChooseCircularWellsRight, GoToAdvanceSettings
from zebrazoom.code.GUI.configFileZebrafish import HeadEmbeded
from zebrazoom.code.GUI.dataAnalysisGUI import CreateExperimentOrganizationExcel, ChooseDataAnalysisMethod, PopulationComparison, BoutClustering, AnalysisOutputFolderClustering
from zebrazoom.code.GUI.rolloverAnalysis import RolloverAnalysis, ManualRolloverClassification
from zebrazoom.code.GUI.troubleshooting import ChooseVideoToTroubleshootSplitVideo, VideoToTroubleshootSplitVideo
import zebrazoom.code.GUI.tracking.customTrackingImplementations


LARGE_FONT= ("Verdana", 12)


def getCurrentResultFolder():
  return currentResultFolder


def excepthook(excType, excValue, traceback_):
  app = QApplication.instance()
  errorMessage = QMessageBox(app.window)
  errorMessage.setIcon(QMessageBox.Icon.Critical)
  errorMessage.setWindowTitle("Error")
  formattedTraceback = traceback.format_exception(excType, excValue, traceback_)
  errorMessage.setText("An error has ocurred: %s" % formattedTraceback[-1])
  informativeText = 'Please report the issue on <a href="https://github.com/oliviermirat/ZebraZoom/issues">Github</a>.'
  if app.configFile and app.savedConfigFile != app.configFile:
    configDir = paths.getConfigurationFolder()
    videoName = os.path.splitext(os.path.basename(app.videoToCreateConfigFileFor))[0]
    configFilename = os.path.join(configDir, '%s_%s_unfinished.json' % (videoName, datetime.now().strftime("%Y_%m_%d-%H_%M_%S")))
    with open(configFilename, 'w') as f:
      json.dump({k: v for k, v in app.configFile.items() if k != "firstFrame" and k != "lastFrame"}, f)
    informativeText += '\nConfig file that was being modified when the error happened was saved to %s.' % configFilename
  errorMessage.setInformativeText(informativeText)
  errorMessage.setDetailedText("    %s" % "    ".join(formattedTraceback))
  errorMessage.setStandardButtons(QMessageBox.StandardButton.Ok | QMessageBox.StandardButton.Cancel)
  errorMessage.setDefaultButton(QMessageBox.StandardButton.Cancel)
  errorMessage.button(QMessageBox.StandardButton.Ok).setText("Continue")
  errorMessage.button(QMessageBox.StandardButton.Cancel).setText("Exit")
  textEdit = errorMessage.findChild(QTextEdit)
  textEdit.setLineWrapMode(QTextEdit.LineWrapMode.NoWrap)
  textEdit.setMarkdown(textEdit.toPlainText())
  layout = errorMessage.layout()
  layout.addItem(QSpacerItem(600, 0), layout.rowCount(), 0, 1, layout.columnCount())
  try:
    if errorMessage.exec() != QMessageBox.StandardButton.Ok:
      sys.exit(1)
  finally:
    sys.__excepthook__(excType, excValue, traceback_)


sys.excepthook = excepthook


class PlainApplication(QApplication):
    def __init__(self, args):
        super().__init__(args)
        if not PYQT6 and sys.platform.startswith('win'):  # qt5 uses deprecated windows API to determine the system font, this works around that issuue
            self.setFont(QApplication.font("QMessageBox"))


class ZebraZoomApp(PlainApplication):
    def __init__(self, args):
        super().__init__(args)

        self.homeDirectory = paths.getRootDataFolder()

        self.savedConfigFile = None
        self.configFile = {}
        self.videoToCreateConfigFileFor = ''
        self.wellLeftBorderX = 0
        self.wellLeftBorderY = 0
        self.headCenterX = 0
        self.headCenterY = 0
        self.organism = ''

        self.configFileHistory = []

        self._ZZoutputLocation = paths.getDefaultZZoutputFolder()
        self._temporaryZZoutputLocation = None

        self.title_font = QFont('Helvetica', 18, QFont.Weight.Bold, True)

        self._busyCursor = False

        self._windows = set()
        self.window = QMainWindow()
        self.window.setWindowIcon(QIcon(os.path.join(sys._MEIPASS if getattr(sys, 'frozen', False) else '', 'icon.ico')))
        self.window.closeEvent = self._windowClosed(self.window, self.window.closeEvent)
        layout = QStackedLayout()
        self.frames = {}
        for idx, F in enumerate((StartPage, VideoToAnalyze, ConfigFilePromp, Patience, ZZoutro, ZZoutroSbatch, SeveralVideos, FolderToAnalyze, EnhanceZZOutput, TailExtremityHE, FolderMultipleROIInitialSelect, ViewParameters, Error, ChooseVideoToCreateConfigFileFor, OptimizeConfigFile, ChooseGeneralExperiment, ChooseCenterOfMassTracking, WellOrganisation, FreelySwimmingExperiment, NbRegionsOfInterest, HomegeneousWellsLayout, CircularOrRectangularWells, NumberOfAnimals, NumberOfAnimals2, NumberOfAnimalsCenterOfMass, IdentifyHeadCenter, IdentifyBodyExtremity, FinishConfig, ChooseCircularWellsLeft, ChooseCircularWellsRight, GoToAdvanceSettings, HeadEmbeded, ChooseDataAnalysisMethod, PopulationComparison, BoutClustering, AnalysisOutputFolderClustering, ChooseVideoToTroubleshootSplitVideo, VideoToTroubleshootSplitVideo, RolloverAnalysis, ManualRolloverClassification)):
            self.frames[F.__name__] = idx
            page = F(self)
            if hasattr(page, 'preferredSize'):
                layout.addWidget(self._wrapWidget(page))
            else:
                layout.addWidget(page)
        central_widget = QWidget(self.window)
        central_widget.setLayout(layout)
        self.window.setWindowTitle('ZebraZoom')
        self.window.setCentralWidget(central_widget)
        layout.currentChanged.connect(self._currentPageChanged)
        self.window.showMaximized()

    def askForZZoutputLocation(self):
        selectedFolder = QFileDialog.getExistingDirectory(self.window, "Select ZZoutput folder", os.path.expanduser("~"))
        if selectedFolder:
          self._ZZoutputLocation = selectedFolder
        # recreate ViewParameters page to ensure it uses the new ZZoutputLocation
        layout = self.window.centralWidget().layout()
        oldWidget = layout.widget(self.frames['ViewParameters'])
        layout.replaceWidget(oldWidget, ViewParameters(self))
        oldWidget.setParent(None)

    @property
    def ZZoutputLocation(self):
        if not self.configFile:
          return self._ZZoutputLocation
        if self._temporaryZZoutputLocation is None:
          self._temporaryZZoutputLocation = tempfile.TemporaryDirectory()
          videoInputsFolder = os.path.join(self._ZZoutputLocation, '.ZebraZoomVideoInputs')
          if os.path.exists(videoInputsFolder):
            shutil.copytree(videoInputsFolder, os.path.join(self._temporaryZZoutputLocation.name, '.ZebraZoomVideoInputs'))
        return self._temporaryZZoutputLocation.name

    def _wrapWidget(self, page):
        page.sizeHint = lambda *args, page=page: QSize(*page.preferredSize)
        wrapperWidget = QWidget()
        wrapperLayout = QVBoxLayout()
        wrapperLayout.addWidget(page, alignment=Qt.AlignmentFlag.AlignCenter)
        wrapperWidget.setLayout(wrapperLayout)
        if hasattr(page, 'setArgs'):
          wrapperWidget.setArgs = page.setArgs
        return wrapperWidget

    def _currentPageChanged(self):
        backBtn = self.window.centralWidget().layout().currentWidget().findChild(QPushButton, "back")
        if backBtn is not None and self.configFileHistory:
            try:
                backBtn.clicked.disconnect()
            except TypeError:  # nothing connected
                pass
            backBtn.clicked.connect(self.configFileHistory[-2])

    def show_frame(self, page_name):
        '''Show a frame for the given page name'''
        if page_name == "StartPage":
            if self.configFile and self.savedConfigFile != self.configFile:
                reply = QMessageBox.question(self.window, "Unsaved Changes",
                                             "Are you sure you want to go back to the start page? Changes made to the config will be lost.",
                                             QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No)
                if reply != QMessageBox.StandardButton.Yes:
                    return True
            self.configFile.clear()
            if self._temporaryZZoutputLocation is not None:
              self._temporaryZZoutputLocation.cleanup()
            self._temporaryZZoutputLocation = None
            self.savedConfigFile = None
            self.videoToCreateConfigFileFor = ''
            self.wellLeftBorderX = 0
            self.wellLeftBorderY = 0
            self.headCenterX = 0
            self.headCenterY = 0
            self.organism = ''
            del self.configFileHistory[:]
        if page_name == "CreateExperimentOrganizationExcel":
          page = CreateExperimentOrganizationExcel(self)
          layout = self.window.centralWidget().layout()
          layout.addWidget(page)
          layout.setCurrentWidget(page)
          def cleanup():
            layout.removeWidget(page)
            layout.currentChanged.disconnect(cleanup)
          layout.currentChanged.connect(cleanup)
        else:
          self.window.centralWidget().layout().setCurrentIndex(self.frames[page_name])

    def _windowClosed(self, window, fn):
        def inner(*args, **kwargs):
            if window is self.window:
                for win in tuple(self._windows):
                    win.close()
                sys.exit(0)
            else:
                if window in self._windows:
                    self._windows.remove(window)
            return fn(*args, **kwargs)
        return inner

    def registerWindow(self, window):
        self._windows.add(window)
        window.closeEvent = self._windowClosed(window, window.closeEvent)

    @contextlib.contextmanager
    def busyCursor(self):
        self.setOverrideCursor(Qt.CursorShape.BusyCursor)
        self._busyCursor = True
        yield
        self.restoreOverrideCursor()
        self._busyCursor = False

    @contextlib.contextmanager
    def suppressBusyCursor(self):
        if self._busyCursor:
            self.restoreOverrideCursor()
        yield
        if self._busyCursor:
            self.setOverrideCursor(Qt.CursorShape.BusyCursor)

    def chooseVideoToAnalyze(self, createValidationVideo, chooseFrames):
        GUI_InitialFunctions.chooseVideoToAnalyze(self, createValidationVideo, chooseFrames)

    def chooseFolderToAnalyze(self, sbatchMode):
        GUI_InitialFunctions.chooseFolderToAnalyze(self, sbatchMode)

    def chooseFolderForTailExtremityHE(self):
        GUI_InitialFunctions.chooseFolderForTailExtremityHE(self)

    def chooseFolderForMultipleROIs(self, askCoordinatesForAll):
        GUI_InitialFunctions.chooseFolderForMultipleROIs(self, askCoordinatesForAll)

    def showViewParameters(self, folder=None):
        self.show_frame("ViewParameters")
        self.window.centralWidget().layout().currentWidget().setFolder(folder)

    def showManualRolloverClassification(self, folder=None):
        self.show_frame("ManualRolloverClassification")
        self.window.centralWidget().layout().currentWidget().setFolder(folder)

    def openConfigurationFileFolder(self, homeDirectory):
        GUI_InitialFunctions.openConfigurationFileFolder(self, homeDirectory)

    def optimizeConfigFile(self):
        self.show_frame("OptimizeConfigFile")
        self.window.centralWidget().layout().currentWidget().refresh()

    def openZZOutputFolder(self, homeDirectory):
        GUI_InitialFunctions.openZZOutputFolder(self, homeDirectory)

    # Config File preparation functions

    def chooseVideoToCreateConfigFileFor(self, controller, reloadConfigFile):
        return configFilePrepareFunctions.chooseVideoToCreateConfigFileFor(self, controller, reloadConfigFile)

    def chooseGeneralExperimentFirstStep(self, controller, freeZebra, headEmbZebra, drosophilia, rodent, other, fastScreen):
        configFilePrepareFunctions.chooseGeneralExperimentFirstStep(self, controller, freeZebra, headEmbZebra, drosophilia, rodent, other, fastScreen)

    def chooseGeneralExperiment(self, controller, freeZebra, headEmbZebra, drosophilia, rodent, other, freeZebra2):
        configFilePrepareFunctions.chooseGeneralExperiment(self, controller, freeZebra, headEmbZebra, drosophilia, rodent, other, freeZebra2)

    def wellOrganisation(self, controller, circular, rectangular, roi, other, multipleROIs, groupSameSizeAndShapeEquallySpacedWells):
        configFilePrepareFunctions.wellOrganisation(self, controller, circular, rectangular, roi, other, multipleROIs, groupSameSizeAndShapeEquallySpacedWells)

    def regionsOfInterest(self, controller, nbwells):
        configFilePrepareFunctions.regionsOfInterest(self, controller, nbwells)

    def homegeneousWellsLayout(self, controller, nbRowsOfWells, nbWellsPerRows, detectBouts):
        configFilePrepareFunctions.homegeneousWellsLayout(self, controller, nbRowsOfWells, nbWellsPerRows, detectBouts)

    def morePreciseFastScreen(self, controller, nbRowsOfWells, nbWellsPerRows, detectBouts):
        configFilePrepareFunctions.morePreciseFastScreen(self, controller, nbRowsOfWells, nbWellsPerRows, detectBouts)

    def circularOrRectangularWells(self, controller, nbRowsOfWells, nbWellsPerRows, nbanimals):
        configFilePrepareFunctions.circularOrRectangularWells(self, controller, nbRowsOfWells, nbWellsPerRows, nbanimals)

    def finishConfig(self, testConfig=False):
        suggestedName = '%s_%s' % (os.path.splitext(os.path.basename(self.videoToCreateConfigFileFor))[0], datetime.now().strftime("%Y_%m_%d-%H_%M_%S"))
        configDir = paths.getConfigurationFolder()
        reference, _ = QFileDialog.getSaveFileName(self.window, "Save config", os.path.join(configDir, suggestedName), "JSON (*.json)")
        if not reference:
          return
        if 'outputFolder' in self.configFile:
          del self.configFile['outputFolder']
        # Ideally would like to remove these four lines below, once the problem with wrong 'firstFrame' and 'lastFrame' being saved in the configuration file is solved
        if "lastFrame" in self.configFile:
          # del self.configFile["lastFrame"]
          self.configFile["lastFrameForBackExtract"] = self.configFile["lastFrame"]
        if "firstFrame" in self.configFile:
          # del self.configFile["firstFrame"]
          self.configFile["firstFrameForBackExtract"] = self.configFile["firstFrame"]

        with open(reference, 'w') as outfile:
          json.dump(self.configFile, outfile)

        self.savedConfigFile = self.configFile.copy()

        if testConfig:
          self.testConfig()
        else:
          self.show_frame("StartPage")

    def chooseCircularWellsLeft(self, controller):
        configFilePrepareFunctions.chooseCircularWellsLeft(self, controller)

    def chooseCircularWellsRight(self, controller):
        configFilePrepareFunctions.chooseCircularWellsRight(self, controller)

    def chooseHeadCenter(self, controller):
        configFilePrepareFunctions.chooseHeadCenter(self, controller)

    def chooseBodyExtremity(self, controller):
        configFilePrepareFunctions.chooseBodyExtremity(self, controller)

    def headEmbededGUI(self, controller, blackBack, whiteBack, noBoutDetect, boutDetection, optionExtendedDescentSearchOption, optionBackgroundExtractionOption):
        configFileZebrafishFunctions.headEmbededGUI(self, controller, blackBack, whiteBack, noBoutDetect, boutDetection, optionExtendedDescentSearchOption, optionBackgroundExtractionOption)

    def detectBouts(self, controller, wellNumber, firstFrameParamAdjust, adjustOnWholeVideo, reloadWellPositions=True):
      adjustParameterInsideAlgoFunctions.detectBouts(self, controller, wellNumber, firstFrameParamAdjust, adjustOnWholeVideo, reloadWellPositions=reloadWellPositions)

    def adjustHeadEmbededTracking(self, controller, wellNumber, firstFrameParamAdjust, adjustOnWholeVideo):
      adjustParameterInsideAlgoFunctions.adjustHeadEmbededTracking(self, controller, wellNumber, firstFrameParamAdjust, adjustOnWholeVideo)

    def adjustFreelySwimTracking(self, controller, wellNumber, firstFrameParamAdjust, adjustOnWholeVideo):
      adjustParameterInsideAlgoFunctions.adjustFreelySwimTracking(self, controller, wellNumber, firstFrameParamAdjust, adjustOnWholeVideo)

    def adjustFreelySwimTrackingAutomaticParameters(self, controller, wellNumber, firstFrameParamAdjust, adjustOnWholeVideo):
      adjustParameterInsideAlgoFunctions.adjustFreelySwimTrackingAutomaticParameters(self, controller, wellNumber, firstFrameParamAdjust, adjustOnWholeVideo)

    def calculateBackground(self, *args, **kwargs):
      adjustParameterInsideAlgoFunctions.calculateBackground(self, *args, **kwargs)

    def calculateBackgroundFreelySwim(self, *args, **kwargs):
      adjustParameterInsideAlgoFunctions.calculateBackgroundFreelySwim(self, *args, **kwargs)

    def goToAdvanceSettings(self, controller, yes, no):
      configFilePrepareFunctions.goToAdvanceSettings(self, controller, yes, no)

    def boutClustering(self, controller, nbClustersToFind, FreelySwimming, HeadEmbeded, minNbBendForBoutDetect=3, nbVideosToSave=0, modelUsedForClustering='', removeOutliers=False, frameStepForDistanceCalculation='4', removeBoutsContainingNanValuesInParametersUsedForClustering=True, forcePandasRecreation=0):
      dataAnalysisGUIFunctions.boutClustering(self, controller, nbClustersToFind, FreelySwimming, HeadEmbeded, minNbBendForBoutDetect, nbVideosToSave, modelUsedForClustering, removeOutliers, frameStepForDistanceCalculation, removeBoutsContainingNanValuesInParametersUsedForClustering, forcePandasRecreation)

    def openAnalysisFolder(self, homeDirectory, specificDirectory):
      dataAnalysisGUIFunctions.openAnalysisFolder(self, homeDirectory, specificDirectory)

    def chooseVideoToTroubleshootSplitVideo(self, controller):
      troubleshootingFunction.chooseVideoToTroubleshootSplitVideo(self, controller)

    def testConfig(self, addToHistory=True):
      videoPath = self.videoToCreateConfigFileFor
      pathToVideo  = os.path.dirname(videoPath)
      videoName, videoExt = os.path.splitext(os.path.basename(videoPath))
      videoExt = videoExt.lstrip('.')

      def callback():
        firstFrame = self.configFile["firstFrame"]
        self.configFile.clear()
        self.configFile.update(configFile)
        lastFrame = min(firstFrame + maximumFramesButtonGroup.checkedId(), int(zzVideoReading.VideoCapture(videoPath).get(7)) - 1)
        del self.configFileHistory[:]
        with self.busyCursor():
          try:
            tabParams = ["zebraZoomVideoAnalysis", pathToVideo, videoName, videoExt, self.configFile,
                         "firstFrame", firstFrame, "lastFrame", lastFrame, "freqAlgoPosFollow", 100,
                         "popUpAlgoFollow", 1, "outputFolder", self.ZZoutputLocation,
                         "backgroundExtractionForceUseAllVideoFrames", int(backgroundExtractionForceUseAllVideoFramesCheckbox.isChecked())]
            ZZanalysis = ZebraZoomVideoAnalysis(pathToVideo, videoName, videoExt, self.configFile, tabParams)
            ZZanalysis.run()
            resultsFilename = ZZanalysis._hyperparameters['videoNameWithTimestamp'] if not ZZanalysis._hyperparameters['storeH5'] else f'{ZZanalysis._hyperparameters["videoNameWithTimestamp"]}.h5'
          except NameError:
            self.show_frame("Error")
            return
          finally:
            self.configFile.clear()
            self.configFile.update(configFile)
        (self.showViewParameters if not addToHistory else util.addToHistory(self.showViewParameters))(resultsFilename)
      configFile = self.configFile.copy()

      layout = QVBoxLayout()
      backgroundExtractionForceUseAllVideoFramesCheckbox = QCheckBox("Use all frames to calculate background")
      backgroundExtractionForceUseAllVideoFramesCheckbox.setChecked(True)
      layout.addWidget(backgroundExtractionForceUseAllVideoFramesCheckbox, alignment=Qt.AlignmentFlag.AlignCenter)
      maximumFramesLayout = QHBoxLayout()
      maximumFramesLayout.addStretch()
      maximumFramesLayout.addWidget(QLabel("Maximum number of frames used for tracking:"))
      maximumFramesButtonGroup = QButtonGroup()
      for value in (50, 100, 200, 300, 500):
        btn = QRadioButton(str(value))
        maximumFramesLayout.addWidget(btn)
        maximumFramesButtonGroup.addButton(btn, id=value)
      maximumFramesButtonGroup.button(500).setChecked(True)
      maximumFramesLayout.addStretch()
      layout.addLayout(maximumFramesLayout)
      cb = util.chooseBeginningPage if not addToHistory else util.addToHistory(util.chooseBeginningPage)
      cb(self, videoPath, "We will test the tracking on the video you provided, with the configuration file you just created. We will do this test only on the selected number of frames (maximum) and the tracking will start at the frame you select below (so please choose a section of the video where the animal is moving).",
         "Ok, I want the tracking to start at this frame!", callback, additionalLayout=layout, titleStyle={'color': 'red', 'font': QFont('Helvetica', 14, QFont.Weight.Bold, True)})



import random
import torch

from torchvision.transforms import functional as F


def _flip_coco_person_keypoints(kps, width):
    flip_inds = [0, 2, 1, 4, 3, 6, 5, 8, 7, 10, 9, 12, 11, 14, 13, 16, 15]
    flipped_data = kps[:, flip_inds]
    flipped_data[..., 0] = width - flipped_data[..., 0]
    # Maintain COCO convention that if visibility == 0, then x, y = 0
    inds = flipped_data[..., 2] == 0
    flipped_data[inds] = 0
    return flipped_data


class Compose(object):
    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, image, target):
        for t in self.transforms:
            image, target = t(image, target)
        return image, target


class RandomHorizontalFlip(object):
    def __init__(self, prob):
        self.prob = prob

    def __call__(self, image, target):
        if random.random() < self.prob:
            height, width = image.shape[-2:]
            image = image.flip(-1)
            bbox = target["boxes"]
            bbox[:, [0, 2]] = width - bbox[:, [2, 0]]
            target["boxes"] = bbox
            if "masks" in target:
                target["masks"] = target["masks"].flip(-1)
            if "keypoints" in target:
                keypoints = target["keypoints"]
                keypoints = _flip_coco_person_keypoints(keypoints, width)
                target["keypoints"] = keypoints
        return image, target


class ToTensor(object):
    def __call__(self, image, target):
        image = F.to_tensor(image)
        return image, target



from zebrazoom.dataAnalysis.dataanalysis.activeLearning import activeLearning

def launchActiveLearning():

  modelUsed     = 'KNeighborsClassifier' # 'SVC'
  nbConditions  = 2
  nbCluster     = 4
  outputFolderResult = 'zebrazoom/dataAnalysis/resultsClustering/parkinsonFreelySwimForClustering_OutliersRemoved_BASSstyleParam2/activeLearning/'
  N_QUERIES     = 40
  manualClassicationPath = 'zebrazoom/dataAnalysis/resultsClustering/parkinsonFreelySwimForClustering_OutliersRemoved_BASSstyleParam2/manualClassifications.txt'

  activeLearning(modelUsed, nbConditions, nbCluster, outputFolderResult, N_QUERIES, manualClassicationPath)



import numpy as np

from ._openResultsFile import openResultsFile


def getNbOscAndTBFPerBoutFromManualClassification(videoName: str, numWell: int, numAnimal: int, numBout: int):
  with openResultsFile(videoName, 'r') as results:
    
    # Retrieving bout data
    if 'videoFPS' not in results.attrs:
      raise ValueError(f'videoFPS is not defined, cannot calculate tail beat frequency')
    fps = results.attrs['videoFPS']
    boutPath     = f'dataForWell{numWell}/dataForAnimal{numAnimal}/listOfBouts/bout{numBout}'
    if boutPath not in results:
      raise ValueError(f"bout {numBout} for animal {numAnimal} in well {numWell} doesn't exist")
    boutGroup = results[boutPath]
    firstFrame = results.attrs['firstFrame']
    start = boutGroup.attrs['BoutStart'] - firstFrame
    end = boutGroup.attrs['BoutEnd'] - firstFrame + 1
    dataGroup = results[f'dataForWell{numWell}/dataForAnimal{numAnimal}/dataPerFrame']
    if 'manualBend' not in dataGroup:
      raise ValueError(f'Manual validation data not found for animal {numAnimal} in well {numWell}')
    
    # Extracting timings of bends
    Bend_Timing = []
    for i in range(start, end):
      if dataGroup['manualBend'][i]:
        Bend_Timing.append(i)
    if not(0 in Bend_Timing):
      Bend_Timing = [start] + Bend_Timing
    
    # Number of Oscillations calculation
    numberOfOscillations = dataGroup['manualBend'][start:end].sum() / 2
    
    # Quotient TBF calculation
    boutDuration = (end - start) / results.attrs['videoFPS']
    quotientTBF = numberOfOscillations / boutDuration
    
    # Instantaneous TBF calculation
    meanOfInstantaneousTBF = float('NaN')
    if len(Bend_Timing):
      meanOfInstantaneousTBF = np.mean(fps / (2 * np.diff(Bend_Timing)))
    
    return numberOfOscillations, quotientTBF, meanOfInstantaneousTBF



import json
import matplotlib.pyplot as plt
import cv2
import numpy as np
import math
from pathlib import Path

import pandas as pd
import os
import shutil
import time
import re

def filterLatencyAndMergeBoutsInSameTrials(nameOfExperiment, minFrameNumberBoutStart, maxFrameNumberBoutStart, calculationMethod, pathToZZoutput, dropDuplicates=False):

  pathToRoot = Path(pathToZZoutput).parent

  dropDuplicatesExtensionName = '_duplicateRemoved' if dropDuplicates else ''
  
  if os.path.exists(os.path.join(pathToRoot, os.path.join('dataAnalysis', os.path.join('resultsKinematic', nameOfExperiment + '_perFish_' + calculationMethod + dropDuplicatesExtensionName)))):
    shutil.rmtree(os.path.join(pathToRoot, os.path.join('dataAnalysis', os.path.join('resultsKinematic', nameOfExperiment + '_perFish_' + calculationMethod + dropDuplicatesExtensionName))))
    time.sleep(0.1)
  
  os.mkdir(os.path.join(pathToRoot, os.path.join('dataAnalysis', os.path.join('resultsKinematic', nameOfExperiment + '_perFish_' + calculationMethod + dropDuplicatesExtensionName))))
  os.mkdir(os.path.join(pathToRoot, os.path.join('dataAnalysis', os.path.join('resultsKinematic', os.path.join(nameOfExperiment + '_perFish_' + calculationMethod + dropDuplicatesExtensionName, 'allBoutsMixed')))))
  os.mkdir(os.path.join(pathToRoot, os.path.join('dataAnalysis', os.path.join('resultsKinematic', os.path.join(nameOfExperiment + '_perFish_' + calculationMethod + dropDuplicatesExtensionName, 'medianPerWellFirst')))))
  
  data = pd.read_excel(os.path.join(pathToRoot, os.path.join('dataAnalysis', os.path.join('resultsKinematic', os.path.join(nameOfExperiment, os.path.join('allBoutsMixed', 'globalParametersInsideCategories.xlsx'))))))
  
  data = data[data['BoutStart'] >= minFrameNumberBoutStart]
  data = data[data['BoutStart'] <= maxFrameNumberBoutStart]
  data = data.reset_index()
  
  if dropDuplicates:
    data = data.drop_duplicates(subset=['Trial_ID', 'Well_ID', 'Genotype', 'Condition'])
    data = data.reset_index()
    del data['level_0']
  
  for i in range(0, len(data)):
    data.loc[i, "Trial_ID"] = re.sub(r"Trial\d+", "", data.loc[i, "Trial_ID"])
  
  if calculationMethod == 'median' or calculationMethod == 'mean':
    data2 = data.copy()
    data2 = data2.groupby(['Trial_ID', 'Well_ID', 'Genotype', 'Condition']).count()
    if calculationMethod == 'median':
      data = data.groupby(['Trial_ID', 'Well_ID', 'Genotype', 'Condition']).median()
    elif calculationMethod == 'mean':
      data = data.groupby(['Trial_ID', 'Well_ID', 'Genotype', 'Condition']).mean()
    del data['index']
    del data['Unnamed: 0']
    data['count'] = data2['index']
  else:
    print("No mean or median applied")
  
  data.to_excel(os.path.join(pathToRoot, os.path.join('dataAnalysis', os.path.join('resultsKinematic', os.path.join(nameOfExperiment + '_perFish_' + calculationMethod + dropDuplicatesExtensionName, os.path.join('allBoutsMixed', 'globalParametersInsideCategories.xlsx'))))))
  
  data.to_csv(os.path.join(pathToRoot, os.path.join('dataAnalysis', os.path.join('resultsKinematic', os.path.join(nameOfExperiment + '_perFish_' + calculationMethod + dropDuplicatesExtensionName, os.path.join('allBoutsMixed', 'globalParametersInsideCategories.csv'))))))
  
  data.to_excel(os.path.join(pathToRoot, os.path.join('dataAnalysis', os.path.join('resultsKinematic', os.path.join(nameOfExperiment + '_perFish_' + calculationMethod + dropDuplicatesExtensionName, os.path.join('medianPerWellFirst', 'globalParametersInsideCategories.xlsx'))))))
  
  data.to_csv(os.path.join(pathToRoot, os.path.join('dataAnalysis', os.path.join('resultsKinematic', os.path.join(nameOfExperiment + '_perFish_' + calculationMethod + dropDuplicatesExtensionName, os.path.join('medianPerWellFirst', 'globalParametersInsideCategories.csv'))))))



import numpy as np
import matplotlib.pyplot as plt

from ._openResultsFile import openResultsFile


def plotManualVsAutomaticBendLocations(videoName: str, numWell: int, numAnimal: int, numBout: int):
  with openResultsFile(videoName, 'r') as results:
    if 'videoFPS' not in results.attrs:
      raise ValueError(f'videoFPS is not defined, cannot calculate tail beat frequency')
    boutPath     = f'dataForWell{numWell}/dataForAnimal{numAnimal}/listOfBouts/bout{numBout}'
    if boutPath not in results:
      raise ValueError(f"bout {numBout} for animal {numAnimal} in well {numWell} doesn't exist")
    boutGroup = results[boutPath]
    firstFrame = results.attrs['firstFrame']
    start = boutGroup.attrs['BoutStart'] - firstFrame
    end = boutGroup.attrs['BoutEnd'] - firstFrame + 1
    dataGroup = results[f'dataForWell{numWell}/dataForAnimal{numAnimal}/dataPerFrame']
    if 'manualBend' not in dataGroup:
      raise ValueError(f'Manual validation data not found for animal {numAnimal} in well {numWell}')
    
    manualBendDetect = dataGroup['manualBend'][start:end]
    manualBendDetect = [1 if manualBendDetect[i] else 0 for i in range(len(manualBendDetect))]
    
    autoBendDetect   = np.array([0 for i in range(len(dataGroup['manualBend'][start:end]))])
    for i in results['dataForWell0/dataForAnimal' + str(numAnimal) + '/listOfBouts/bout' + str(numBout) + '/Bend_Timing'][:]:
      if i < len(autoBendDetect):
        autoBendDetect[i] = 2
    
    plt.plot(manualBendDetect, label='Manual Bend Detect')
    plt.plot(autoBendDetect,   label='Automatic Bend Detect')
    
    plt.legend()
    
    plt.show()
    


from PyQt5.QtCore import Qt
from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import QLabel, QWidget, QPushButton, QHBoxLayout, QVBoxLayout, QRadioButton, QButtonGroup

import zebrazoom.code.util as util


class HeadEmbeded(QWidget):
  def __init__(self, controller):
    super().__init__(controller.window)
    self.controller = controller
    self.preferredSize = (1152, 768)

    layout = QVBoxLayout()
    layout.addWidget(util.apply_style(QLabel("Prepare Config File", self), font=controller.title_font), alignment=Qt.AlignmentFlag.AlignCenter)

    layout.addWidget(util.apply_style(QLabel("Choose only one of the options below:", self), font=QFont("Helvetica", 12)), alignment=Qt.AlignmentFlag.AlignCenter)
    btnGroup1 = QButtonGroup(self)
    blackBackRadioButton = QRadioButton("Black background, white zebrafish.", self)
    btnGroup1.addButton(blackBackRadioButton)
    blackBackRadioButton.setChecked(True)
    layout.addWidget(blackBackRadioButton, alignment=Qt.AlignmentFlag.AlignCenter)
    whiteBackRadioButton = QRadioButton("White background, black zebrafish.", self)
    btnGroup1.addButton(whiteBackRadioButton)
    layout.addWidget(whiteBackRadioButton, alignment=Qt.AlignmentFlag.AlignCenter)

    layout.addWidget(util.apply_style(QLabel("Do you want ZebraZoom to detect bouts of movement?", self), font=QFont("Helvetica", 12)), alignment=Qt.AlignmentFlag.AlignCenter)
    btnGroup2 = QButtonGroup(self)
    noBoutDetectRadioButton = QRadioButton("No. I want the tracking data for all frames of the videos.", self)
    btnGroup2.addButton(noBoutDetectRadioButton)
    noBoutDetectRadioButton.setChecked(True)
    layout.addWidget(noBoutDetectRadioButton, alignment=Qt.AlignmentFlag.AlignCenter)
    boutDetectionRadioButton = QRadioButton("Yes. I want the tracking data only when the fish is moving.", self)
    btnGroup2.addButton(boutDetectionRadioButton)
    layout.addWidget(boutDetectionRadioButton, alignment=Qt.AlignmentFlag.AlignCenter)

    layout.addWidget(util.apply_style(QLabel("Do you want to try to tweak tracking parameters further?", self), font=QFont("Helvetica", 12)), alignment=Qt.AlignmentFlag.AlignCenter)
    layout.addWidget(QLabel("Warning: further tweaking tracking parameters could make tracking results worse; please try without this option first.", self), alignment=Qt.AlignmentFlag.AlignCenter)
    btnGroup3 = QButtonGroup(self)
    tweakTrackingParamsYesRadioButton = QRadioButton("Yes", self)
    btnGroup3.addButton(tweakTrackingParamsYesRadioButton)
    tweakTrackingParamsYesRadioButton.setChecked(True)
    layout.addWidget(tweakTrackingParamsYesRadioButton, alignment=Qt.AlignmentFlag.AlignCenter)
    tweakTrackingParamsNoRadioButton = QRadioButton("No", self)
    btnGroup3.addButton(tweakTrackingParamsNoRadioButton)
    layout.addWidget(tweakTrackingParamsNoRadioButton, alignment=Qt.AlignmentFlag.AlignCenter)

    buttonsLayout = QHBoxLayout()
    buttonsLayout.addStretch()
    backBtn = QPushButton("Back", self)
    backBtn.setObjectName("back")
    buttonsLayout.addWidget(backBtn, alignment=Qt.AlignmentFlag.AlignCenter)
    startPageBtn = QPushButton("Go to the start page", self)
    startPageBtn.clicked.connect(lambda: controller.show_frame("StartPage"))
    buttonsLayout.addWidget(startPageBtn, alignment=Qt.AlignmentFlag.AlignCenter)
    nextBtn = util.apply_style(QPushButton("Next", self), background_color=util.DEFAULT_BUTTON_COLOR)
    nextBtn.clicked.connect(lambda: controller.headEmbededGUI(controller, blackBackRadioButton.isChecked(), whiteBackRadioButton.isChecked(), noBoutDetectRadioButton.isChecked(), boutDetectionRadioButton.isChecked(), tweakTrackingParamsYesRadioButton.isChecked(), tweakTrackingParamsNoRadioButton.isChecked()))
    buttonsLayout.addWidget(nextBtn, alignment=Qt.AlignmentFlag.AlignCenter)
    buttonsLayout.addStretch()
    layout.addLayout(buttonsLayout)

    self.setLayout(layout)



from ._openResultsFile import openResultsFile


def listAllBouts(videoName: str, numWell: int, numAnimal: int, seconds: bool=False):
  with openResultsFile(videoName, 'r') as results:
    boutsPath = f'dataForWell{numWell}/dataForAnimal{numAnimal}/listOfBouts'
    if boutsPath not in results:
      raise ValueError(f"bouts not found for animal {numAnimal} in well {numWell}")
    boutsGroup = results[boutsPath]
    boutTimings = ((boutsGroup[f'bout{boutIdx}'].attrs['BoutStart'], boutsGroup[f'bout{boutIdx}'].attrs['BoutEnd']) for boutIdx in range(boutsGroup.attrs['numberOfBouts']))
    if not seconds:
      return list(boutTimings)
    if 'videoFPS' not in results.attrs:
      raise ValueError(f'videoFPS not found in the results, cannot convert frames to seconds')
    videoFPS = results.attrs['videoFPS']
    return [(start / videoFPS, end / videoFPS) for start, end in boutTimings]



import math
import pickle
import os
import cv2

import h5py
import numpy as np

from zebrazoom.code.dataPostProcessing.dataPostProcessing import dataPostProcessing

from ._base import BaseTrackingMethod
from ._getBackground import GetBackgroundMixin
from ._updateBackgroundAtInterval import UpdateBackgroundAtIntervalMixin


class BaseZebraZoomTrackingMethod(BaseTrackingMethod, GetBackgroundMixin, UpdateBackgroundAtIntervalMixin):
  '''Base class for tracking implementations which use ZebraZoom mixins.'''

  def _debugFrame(self, frame, title=None, buttons=(), timeout=None):
    raise ValueError("debug mode is not available without the GUI")

  def dataPostProcessing(self, outputFolder, superStruct):
    # Various post-processing options depending on configuration file choices
    return dataPostProcessing(outputFolder, superStruct, self._hyperparameters, self._hyperparameters['videoNameWithTimestamp'], os.path.splitext(os.path.basename(self._videoPath))[1])

  def _debugTracking(self, frameNumber: int, output: list, outputHeading: list, frame2: np.array) -> None:
    pass

  @staticmethod
  def __rollingMedianFilter(array, window):
    array2 = np.convolve(array, np.ones(window), 'same') / window
    array2[:window-1] = array[:window-1]
    array2[-window+1:] = array[-window+1:]
    return array2
  
  def _postProcessHeadingWithTrajectory(self, trackingHeadTailAllAnimals, trackingHeadingAllAnimals, trackingProbabilityOfHeadingGoodCalculation):
    for animalId in range(0, len(trackingHeadTailAllAnimals)):
      last_index   = 0
      last_xHead   = trackingHeadTailAllAnimals[animalId][last_index][0][0]
      last_yHead   = trackingHeadTailAllAnimals[animalId][last_index][0][1]
      last_heading = (trackingHeadingAllAnimals[animalId][last_index] + math.pi) % (2*math.pi)
      curDist = 0
      for frameNumber in range(0, len(trackingHeadTailAllAnimals[animalId])):
        xHead = trackingHeadTailAllAnimals[animalId][frameNumber][0][0]
        yHead = trackingHeadTailAllAnimals[animalId][frameNumber][0][1]
        heading = trackingHeadingAllAnimals[animalId][frameNumber]
        dist = math.sqrt((xHead - last_xHead)**2 + (yHead - last_yHead)**2)
        if dist > self._hyperparameters["postProcessHeadingWithTrajectory_minDist"]:
          last_heading = self._calculateAngle(xHead, yHead, last_xHead, last_yHead)
          last_index += 1
          last_xHead   = trackingHeadTailAllAnimals[animalId][last_index][0][0]
          last_yHead   = trackingHeadTailAllAnimals[animalId][last_index][0][1]
        if self._distBetweenThetas((heading + math.pi) % (2*math.pi), last_heading) > self._distBetweenThetas(heading, last_heading) and self._distBetweenThetas(heading, last_heading) < 0.25 * math.pi and abs(trackingProbabilityOfHeadingGoodCalculation[0, frameNumber]) <= 10:
          trackingHeadingAllAnimals[animalId][frameNumber] = (heading + math.pi) % (2*math.pi)
          print("Inversion of heading in post processing, animalId:", animalId, " ; frame:", frameNumber, "; proba:", trackingProbabilityOfHeadingGoodCalculation[0, frameNumber])
      
      potentiallyAbnormalRangeIndStart = -1
      for revert in [0, 1]:
        if revert:
          trackingHeadingAllAnimals[animalId] = np.flip(trackingHeadingAllAnimals[animalId])
        for frameNumber in range(1, len(trackingHeadTailAllAnimals[animalId])):
          heading = trackingHeadingAllAnimals[animalId][frameNumber]
          if self._distBetweenThetas(heading, trackingHeadingAllAnimals[animalId][potentiallyAbnormalRangeIndStart-1 if potentiallyAbnormalRangeIndStart!=-1 else frameNumber-1]) > 0.8 * math.pi:
            if potentiallyAbnormalRangeIndStart == -1:
              potentiallyAbnormalRangeIndStart = frameNumber
          else:
            if potentiallyAbnormalRangeIndStart != -1:
              if frameNumber - potentiallyAbnormalRangeIndStart > 100:
                print("heading post processing second step: potentiallyAbnormalRange too long, reseting")
                potentiallyAbnormalRangeIndStart = -1
              else:
                print("Second heading post processing step: looking for potential invertion between frame", potentiallyAbnormalRangeIndStart, "and", frameNumber)
                for frameNumber2 in range(potentiallyAbnormalRangeIndStart, frameNumber):
                  if self._distBetweenThetas(trackingHeadingAllAnimals[animalId][frameNumber2], trackingHeadingAllAnimals[animalId][potentiallyAbnormalRangeIndStart-1]) > math.pi / 2:
                    print("Second heading post processing step: inverting angle for frame", frameNumber2)
                    trackingHeadingAllAnimals[animalId][frameNumber2] = (trackingHeadingAllAnimals[animalId][frameNumber2] + math.pi) % (2*math.pi)
                potentiallyAbnormalRangeIndStart = -1
      trackingHeadingAllAnimals[animalId] = np.flip(trackingHeadingAllAnimals[animalId])
      
      for frameNumber in range(1, len(trackingHeadTailAllAnimals[animalId])-1):
        headingBef  = trackingHeadingAllAnimals[animalId][frameNumber-1]
        headingPres = trackingHeadingAllAnimals[animalId][frameNumber]
        headingAft  = trackingHeadingAllAnimals[animalId][frameNumber+1]
        if self._distBetweenThetas(headingBef, headingPres) > 110*(math.pi/180) and self._distBetweenThetas(headingPres, headingAft) > 110*(math.pi/180):
          print("Third heading post processing step: inverting frame", frameNumber)
          trackingHeadingAllAnimals[animalId][frameNumber] = (trackingHeadingAllAnimals[animalId][frameNumber] + math.pi) % (2 * math.pi)

  
  def _postProcessMultipleTrajectories(self, trackingHeadTailAllAnimals, trackingProbabilityOfGoodDetection):
    maxDistanceAuthorized = self._hyperparameters["postProcessMaxDistanceAuthorized"]
    maxDisapearanceFrames = self._hyperparameters["postProcessMaxDisapearanceFrames"]

    # Removing all points for which the sum of all pixels of the inverted image (the "probability of good detection") is too low
    if self._hyperparameters["postProcessRemoveLowProbabilityDetection"]:
      for animalId in range(0, len(trackingHeadTailAllAnimals)):
        probabilityOfGoodDetection = trackingProbabilityOfGoodDetection[animalId, :]
        if self._hyperparameters["postProcessLowProbabilityDetectionPercentOfMaximum"] == 0:
          toRemove   = (probabilityOfGoodDetection - np.mean(probabilityOfGoodDetection) < -self._hyperparameters["postProcessLowProbabilityDetectionThreshold"] * np.std(probabilityOfGoodDetection))
        else:
          toRemove   = (probabilityOfGoodDetection < np.max(probabilityOfGoodDetection) * self._hyperparameters["postProcessLowProbabilityDetectionPercentOfMaximum"])
        print("Well ", animalId, ": Number of elements to remove:", np.sum(toRemove), " out of ", len(toRemove), "elements")
        trackingHeadTailAllAnimals[animalId, toRemove, 0, 0] = 0
        trackingHeadTailAllAnimals[animalId, toRemove, 0, 1] = 0

    # Removing all points that are too close to the borders
    if self._hyperparameters["postProcessRemovePointsOnBordersMargin"]:
      borderMargin = self._hyperparameters["postProcessRemovePointsOnBordersMargin"]
      for animalId in range(0, len(trackingHeadTailAllAnimals)):
        for frameNumber in range(0, len(trackingHeadTailAllAnimals[animalId])):
          xHead = trackingHeadTailAllAnimals[animalId][frameNumber][0][0]
          yHead = trackingHeadTailAllAnimals[animalId][frameNumber][0][1]
          if (xHead <= borderMargin) or (yHead <= borderMargin) or (xHead >= self._wellPositions[animalId]["lengthX"] - borderMargin - 1) or (yHead >= self._wellPositions[animalId]["lengthY"] - borderMargin - 1):
            trackingHeadTailAllAnimals[animalId][frameNumber][0][0] = 0
            trackingHeadTailAllAnimals[animalId][frameNumber][0][1] = 0

    # Removing all points that are deviating too much from the main trajectory
    if self._hyperparameters["postProcessRemovePointsAwayFromMainTrajectory"]:
      for animalId in range(0, len(trackingHeadTailAllAnimals)):
        xHeadPositions = trackingHeadTailAllAnimals[animalId, :, 0, 0]
        yHeadPositions = trackingHeadTailAllAnimals[animalId, :, 0, 1]
        xHeadPositionsRollingMedian = self.__rollingMedianFilter(xHeadPositions, 11)
        yHeadPositionsRollingMedian = self.__rollingMedianFilter(yHeadPositions, 11)
        distance = np.sqrt((xHeadPositions - xHeadPositionsRollingMedian) ** 2 + (yHeadPositions - yHeadPositionsRollingMedian) ** 2)
        distanceRollingMedian = self.__rollingMedianFilter(distance, 5)
        normalizedDistance = np.array([1 for x in range(0,len(xHeadPositions))])
        normalizedDistance = distance / distanceRollingMedian
        normalizedDistance = np.nan_to_num(normalizedDistance, nan=1)
        toRemove   = normalizedDistance - np.mean(normalizedDistance) > self._hyperparameters["postProcessRemovePointsAwayFromMainTrajectoryThreshold"] * np.std(normalizedDistance)
        trackingHeadTailAllAnimals[animalId, toRemove, 0, 0] = 0
        trackingHeadTailAllAnimals[animalId, toRemove, 0, 1] = 0

    currentlyZero  = False
    zeroFrameStart = 0
    for animalId in range(0, len(trackingHeadTailAllAnimals)):
      for frameNumber in range(0, len(trackingHeadTailAllAnimals[animalId])):
        xHead = trackingHeadTailAllAnimals[animalId][frameNumber][0][0]
        yHead = trackingHeadTailAllAnimals[animalId][frameNumber][0][1]
        if frameNumber != 0:
          xHeadPrev = trackingHeadTailAllAnimals[animalId][frameNumber-1][0][0]
          yHeadPrev = trackingHeadTailAllAnimals[animalId][frameNumber-1][0][1]
        else:
          xHeadPrev = trackingHeadTailAllAnimals[animalId][frameNumber][0][0]
          yHeadPrev = trackingHeadTailAllAnimals[animalId][frameNumber][0][1]
        if ((xHead == 0 and yHead == 0) or (math.sqrt((xHead - xHeadPrev)**2 + (yHead - yHeadPrev)**2) > maxDistanceAuthorized)) and (frameNumber != len(trackingHeadTailAllAnimals[animalId]) - 1):
          if not(currentlyZero):
            zeroFrameStart = frameNumber
          # print("currentlyZero at True: animalId:", animalId, "; frameNumber:", frameNumber)
          currentlyZero = True
        else:
          if currentlyZero:
            if zeroFrameStart >= 1:
              xHeadStart = trackingHeadTailAllAnimals[animalId][zeroFrameStart-1][0][0]
              yHeadStart = trackingHeadTailAllAnimals[animalId][zeroFrameStart-1][0][1]
            else:
              xHeadStart = trackingHeadTailAllAnimals[animalId][frameNumber][0][0]
              yHeadStart = trackingHeadTailAllAnimals[animalId][frameNumber][0][1]
            xHeadEnd = trackingHeadTailAllAnimals[animalId][frameNumber][0][0]
            yHeadEnd = trackingHeadTailAllAnimals[animalId][frameNumber][0][1]
            if (math.sqrt((xHeadEnd - xHeadStart)**2 + (yHeadEnd - yHeadStart)**2) < maxDistanceAuthorized) or (frameNumber - zeroFrameStart > maxDisapearanceFrames):
              currentlyZero = False
              if (math.sqrt((xHeadEnd - xHeadStart)**2 + (yHeadEnd - yHeadStart)**2) < maxDistanceAuthorized) and not((xHeadEnd == 0) and (yHeadEnd == 0)):
                xStep = (xHeadEnd - xHeadStart) / (frameNumber - zeroFrameStart)
                yStep = (yHeadEnd - yHeadStart) / (frameNumber - zeroFrameStart)
                for frameAtZeroToChange in range(zeroFrameStart, frameNumber):
                  trackingHeadTailAllAnimals[animalId][frameAtZeroToChange][0][0] = xHeadStart + xStep * (frameAtZeroToChange - zeroFrameStart)
                  trackingHeadTailAllAnimals[animalId][frameAtZeroToChange][0][1] = yHeadStart + yStep * (frameAtZeroToChange - zeroFrameStart)
              else:
                for frameAtZeroToChange in range(zeroFrameStart, frameNumber):
                  trackingHeadTailAllAnimals[animalId][frameAtZeroToChange][0][0] = xHeadStart
                  trackingHeadTailAllAnimals[animalId][frameAtZeroToChange][0][1] = yHeadStart

  @staticmethod
  def _calculateAngle(xStart, yStart, xEnd, yEnd):
    vx = xEnd - xStart
    vy = yEnd - yStart
    if vx == 0:
      if vy > 0:
        lastFirstTheta = math.pi/2
      else:
        lastFirstTheta = (3*math.pi)/2
    else:
      lastFirstTheta = np.arctan(abs(vy/vx))
      if (vx < 0) and (vy >= 0):
        lastFirstTheta = math.pi - lastFirstTheta
      elif (vx < 0) and (vy <= 0):
        lastFirstTheta = lastFirstTheta + math.pi
      elif (vx > 0) and (vy <= 0):
        lastFirstTheta = 2*math.pi - lastFirstTheta
    return lastFirstTheta

  @staticmethod
  def _distBetweenThetas(theta1, theta2):
    diff = 0
    if theta1 > theta2:
      diff = theta1 - theta2
    else:
      diff = theta2 - theta1
    if diff > math.pi:
      diff = (2 * math.pi) - diff
    return diff

  @staticmethod
  def _assignValueIfBetweenRange(value, minn, maxx):
    if value < minn:
      return minn
    if value > maxx:
      return maxx
    return value

  def getBackground(self):
    if self._hyperparameters["backgroundSubtractorKNN"] or (self._hyperparameters["headEmbeded"] and self._hyperparameters["headEmbededRemoveBack"] == 0 and self._hyperparameters["headEmbededAutoSet_BackgroundExtractionOption"] == 0 and not self._hyperparameters['adjustHeadEmbeddedEyeTracking'] and self._hyperparameters["adjustHeadEmbededTracking"] == 0) or self._hyperparameters["trackingDL"] or self._hyperparameters["fishTailTrackingDifficultBackground"]:
      background = []
    else:
      print("start get background")
      if self._hyperparameters["reloadBackground"]:
        fname = next(reversed(sorted(name for name in os.listdir(self._hyperparameters['outputFolder']) if os.path.splitext(name)[0][:-20] == self._videoName and os.path.splitext(name)[0] != self._hyperparameters['videoNameWithTimestamp'])))
        with h5py.File(os.path.join(self._hyperparameters['outputFolder'], fname)) as results:
          background = results['background'][:]
      else:
        background = self._getBackground()
      if self._hyperparameters['storeH5']:
        with h5py.File(self._hyperparameters['H5filename'], 'a') as results:
          results.create_dataset('background', data=background)
      else:
        outputFolderVideo = os.path.join(self._hyperparameters['outputFolder'], self._hyperparameters['videoNameWithTimestamp'])
        if not os.path.exists(outputFolderVideo):
          os.makedirs(outputFolderVideo)
        cv2.imwrite(os.path.join(outputFolderVideo, 'background.png'), background)

    if self._hyperparameters["exitAfterBackgroundExtraction"]:
      print("exitAfterBackgroundExtraction")
      raise ValueError

    return background



import math
import os

import cv2
import numpy as np
import pandas as pd

from .plotSingleFrameTrackingPoints import plotSingleFrameTrackingPoints
from .getCurvaturePerBout import getCurvaturePerBout
from .getFPSandPixelSize import getFPSandPixelSize
from .getDataPerBout import getDataPerBout
from .getSingleFrameDistanceToAxis import getSingleFrameDistanceToAxis


def getSingleFrameTrackingAndDistanceToAxisForFolder(pathToFolder: str):
  for basename in os.listdir(pathToFolder):
    if not basename.endswith('.h5'):
      continue
    filename = os.path.join(pathToFolder, basename)
    outputFolder = os.path.join(pathToFolder, 'output')
    if not os.path.exists(outputFolder):
      os.mkdir(outputFolder)
    frame = plotSingleFrameTrackingPoints(filename)
    # plot the main axis
    heading = getDataPerBout(filename, 0, 0, 0, 'Heading')[0]
    x, y = getDataPerBout(filename, 0, 0, 0, 'HeadPos')[0]
    length = max(frame.shape[:2])  # ensure the line is plotted over the whole frame
    xOffset = length * math.cos(heading)
    yOffset = length * math.sin(heading)
    thickness = 2
    cv2.line(frame, (round(x - xOffset), round(y - yOffset)), (round(x + xOffset), round(y + yOffset)), (0, 0, 255), thickness)
    cv2.imwrite(os.path.join(outputFolder, basename.replace('.h5', '.png')), frame)

    _, pixelSize = getFPSandPixelSize(filename)
    curvature, xTimeValues, yDistanceAlongTheTail = getCurvaturePerBout(filename, 0, 0, 0)
    TailPosX = getDataPerBout(filename, 0, 0, 0, 'TailPosX')
    TailPosY = getDataPerBout(filename, 0, 0, 0, 'TailPosY')
    pd.DataFrame(np.column_stack((np.insert(yDistanceAlongTheTail[:, 0][::-1], 0, 0), TailPosX[0, :], TailPosY[0, :], getSingleFrameDistanceToAxis(filename, pixelSize))), columns=['distanceAlongTheTail', 'TailPosX', 'TailPosY', 'distanceToAxis']).to_excel(os.path.join(outputFolder, basename.replace('.h5', '.xlsx')))



from zebrazoom.code.dataPostProcessing.createPandasDataFrameOfParameters import createPandasDataFrameOfParameters
from zebrazoom.code.getHyperparameters import getHyperparametersSimple

from ._createSuperStructFromH5 import createSuperStructFromH5
from ._openResultsFile import openResultsFile

import matplotlib.pyplot as plt
import numpy as np
import math

def remove_outliers(data, threshold=3):
  mean_value = sum(data) / len(data)
  std_dev = (sum((x - mean_value) ** 2 for x in data) / len(data)) ** 0.5
  # Define a range beyond which values are considered outliers
  lower_bound = mean_value - threshold * std_dev
  upper_bound = mean_value + threshold * std_dev
  # Filter out values outside the defined range
  filtered_data = [x for x in data if lower_bound <= x <= upper_bound]
  return filtered_data


def plotKinematicParametersHist(videoName: str, kinematicParameterName: str, outlierRemoval=False) -> dict:
  
  parameter = []
  with openResultsFile(videoName, 'r+') as results:
    for numWell in range(len(results['wellPositions'])):
      for numAnimal in range(len(results['dataForWell' + str(numWell)])):
        animalPath = f'dataForWell{numWell}/dataForAnimal{numAnimal}'
        if animalPath not in results:
          raise ValueError(f"data for animal {numAnimal} in well {numWell} doesn't exist")
        animalGroup = results[animalPath]
        dataset = animalGroup['kinematicParametersPerBout']
        numberOfBouts, = dataset.shape
        for numBout in range(numberOfBouts):
          parameter.append(dataset[kinematicParameterName][numBout])
  
  if outlierRemoval:
    plt.hist(remove_outliers(parameter), bins=50)
    hist, edges = np.histogram([x for x in parameter if not math.isnan(x)], bins=50)
  else:
    plt.hist(parameter, bins=50)
    hist, edges = np.histogram([x for x in parameter if not math.isnan(x)], bins=50)
  
  max_bin_index = np.argmax(hist)
  max_bin_range = (edges[max_bin_index], edges[max_bin_index + 1])
  print(kinematicParameterName, "; Max bin range:", max_bin_range)
  # print("Median:", np.median([x for x in parameter if not math.isnan(x)]))
  # print("Mean:", np.mean([x for x in parameter if not math.isnan(x)]))
  
  plt.title(kinematicParameterName)
  plt.show()
  
  return parameter, max_bin_range
  


import cv2
import numpy as np
import os
import pickle
import webbrowser


MAX_INT32 = 2 ** 31 - 1


def __calculateSteps(maxDepth, numberOfSteps):
  steps = [maxDepth / 1.8, maxDepth / 2.4, maxDepth / 3.6]
  if numberOfSteps <= 3:
    return reversed(steps[:numberOfSteps])
  if numberOfSteps > 3:
    steps.extend((maxDepth / (3.6 + 0.2 * (numberOfSteps - i)) for i in reversed(range(numberOfSteps - 3))))
  return reversed(steps)


def __createComboBox(layout, status, values, info, names, widgets, nameIdx):
  from PyQt5.QtWidgets import QComboBox, QGridLayout, QWidget

  name, choices, callback, getCurrentIndex = info
  combobox = QComboBox()
  combobox.setContentsMargins(20, 5, 20, 5)
  combobox.addItems(choices)
  combobox.setCurrentIndex(getCurrentIndex(values))
  listView = combobox.view()
  listView.setWordWrap(True)
  listView.adjustSize()
  combobox.currentIndexChanged.connect(lambda idx: callback(idx, values))
  combobox.currentIndexChanged.connect(lambda idx: widgets['loop'].exit())

  layout = QGridLayout()  # this is required to align combobox with sliders
  layout.setColumnStretch(0, 1)
  layout.setRowStretch(0, 1)
  layout.setColumnStretch(2, 1)
  layout.setVerticalSpacing(0)
  layout.setContentsMargins(20, 5, 20, 5)
  layout.addWidget(combobox, 1, 1)
  wrapperWidget = QWidget()
  wrapperWidget.setLayout(layout)

  return name, wrapperWidget


def __createSlider(layout, status, values, info, names, widgets, hasCheckbox, nameIdx, widgetPosition=None):
  from PyQt5.QtCore import QAbstractListModel

  import zebrazoom.code.util as util

  stepIdx = None
  if isinstance(names, QAbstractListModel):
    stepsIdx = names.itemlist.index('steps')
    stepCount = len(values['steps'])
    names.updateSteps(stepCount)
    if stepsIdx < nameIdx <= stepsIdx + stepCount:
      stepIdx = nameIdx - (stepsIdx + 1)
      nameIdx = stepsIdx
      name = f'Step {stepIdx + 1}'
    else:
      name = names.itemlist[nameIdx]
      if nameIdx > stepsIdx:
        nameIdx -= stepCount
    minn, maxx, hint = info[nameIdx]
  else:
    name = names
    minn, maxx, hint = info
  if name not in values:
    assert name.startswith('Step')
    value = values['steps'][stepIdx]
    minn = 0.01
    maxx = max(values['maxDepth'], value * 1.33)
  else:
    value = values[name] if name != 'steps' else len(values[name])
    if value < minn:
      minn = int(value)
    if value > maxx:
      maxx = int(value * 1.33)
  double = name in ("authorizedRelativeLengthTailEnd", "authorizedRelativeLengthTailEnd2", "maxDepth", "thetaDiffAccept", "thetaDiffAcceptAfterAuthorizedRelativeLengthTailEnd", "thetaDiffAcceptAfterAuthorizedRelativeLengthTailEnd2") or name not in values
  if isinstance(names, QAbstractListModel):
    slider = util.SliderWithSpinbox(value, minn, maxx, name=name, double=double, choices=names)

    def choiceChanged(idx):
      if not slider.isVisible():
        return
      choiceIdx = idx // 2
      if names.itemlist[choiceIdx] in widgets:
        widget = widgets[names.itemlist[choiceIdx]]
        blocked = widget.blockSignals(True)
        widget.setChoice(idx)
        widget.blockSignals(blocked)
        layout.replaceWidget(slider, widget)
        slider.hide()
        widget.show()
      else:
        layout.removeWidget(slider)
        slider.hide()
        _createWidget(layout, status, values, info, names, widgets, hasCheckbox, nameIdx=choiceIdx, widgetPosition=widgetPosition)
    slider.choiceChanged.connect(choiceChanged)
  else:
    slider = util.SliderWithSpinbox(value, minn, maxx, name=name, double=double)
  if name == "eyeFilterKernelSize" or name == "paramGaussianBlur":
    slider.setSingleStep(2)

  def showHint(fn):
    def inner(evt):
      status.setText(hint)
      return fn(evt)
    return inner
  slider.enterEvent = showHint(slider.enterEvent)

  def hideHint(fn):
    def inner(evt):
      status.setText(None)
      return fn(evt)
    return inner
  slider.leaveEvent = hideHint(slider.leaveEvent)

  def valueChanged():
    if name == 'paramGaussianBlur' and not slider.value() % 2:
      slider.setValue(slider.value() + 1)
      return
    if name == 'paramGaussianBlurForHeadPosition' and slider.value() and not slider.value() % 2:
      slider.setValue(slider.value() + 1)
      return
    if name not in values:
      assert name.startswith('Step ')
      values['steps'][stepIdx] = slider.value()
    elif name == 'steps':
      newSteps = slider.value()
      currentSteps = len(values[name])
      difference = newSteps - currentSteps
      if difference > 0:
        newValues = []
        for idx in range(difference):
          stepName = f'Step {idx + 1}'
          defaultSteps = list(__calculateSteps(values['maxDepth'], newSteps))
          newValues.append(defaultSteps[idx])
        values[name][:0] = newValues
      elif difference < 0:
        del values[name][:-difference]
      names.updateSteps(len(values[name]))
    else:
      values[name] = slider.value()
    if name == 'headEmbededParamTailDescentPixThreshStop':
      values['maximumMedianValueOfAllPointsAlongTheTail'] = slider.value()
      values['minimumHeadPixelValue'] = slider.value()
    widgets['loop'].exit()
  slider.valueChanged.connect(valueChanged)
  return name, slider

def _createWidget(layout, status, values, info, names, widgets, hasCheckbox, nameIdx=0, widgetPosition=None):
  from PyQt5.QtCore import Qt

  if isinstance(names, tuple):
    name, widget = __createComboBox(layout, status, values, info, names, widgets, nameIdx)
  else:
    name, widget = __createSlider(layout, status, values, info, names, widgets, hasCheckbox, nameIdx, widgetPosition=layout.count()if widgetPosition is None else widgetPosition)

  if name != "Frame number":
    elements = (widgetPosition if widgetPosition is not None else layout.count()) - (4 if hasCheckbox else 3)  # frame, status, checkbox, frameSlider
    row = elements // 2 + (4 if hasCheckbox else 3)
    col = elements % 2
    if len(values) == 1:
      layout.addWidget(widget, row, col, 1, 2, Qt.AlignmentFlag.AlignCenter)
    else:
      layout.addWidget(widget, row, col, Qt.AlignmentFlag.AlignLeft if col else Qt.AlignmentFlag.AlignRight)
    widgets[name] = widget
  else:
    widgets['frameSlider'] = widget
    layout.addWidget(widget, 3 if hasCheckbox else 2, 0, 1, 2, Qt.AlignmentFlag.AlignCenter)


def adjustHyperparameters(l, hyperparameters, hyperparametersListNames, frameToShow, title, organizationTab, widgets, documentationLink=None, addContrastCheckbox=False, addZoomCheckbox=False, addUnprocessedFrameCheckbox=False):
  from PyQt5.QtCore import Qt, QAbstractListModel, QEventLoop, QTimer
  from PyQt5.QtGui import QCursor
  from PyQt5.QtWidgets import QApplication, QCheckBox, QGridLayout, QHBoxLayout, QLabel, QPushButton, QSpinBox, QVBoxLayout, QWidget

  import zebrazoom.code.paths as paths
  import zebrazoom.code.util as util

  app = QApplication.instance()
  stackedLayout = app.window.centralWidget().layout()
  timers = []

  hasCheckbox = addContrastCheckbox or addZoomCheckbox or addUnprocessedFrameCheckbox

  if widgets is None:
    widgets = {'Frame number': l, 'saved': False, 'discarded': False, 'loop': QEventLoop()}
    layout = QGridLayout()
    layout.setVerticalSpacing(0)
    frame = QLabel()
    widgets['frame'] = frame
    layout.addWidget(frame, 0, 0, 1, 2, Qt.AlignmentFlag.AlignCenter)
    layout.setRowStretch(0, 1)
    status = QLabel()
    layout.addWidget(status, 1, 0, 1, 2, Qt.AlignmentFlag.AlignCenter)

    if hasCheckbox:
      checkboxesLayout = QHBoxLayout()
      checkboxesLayout.addStretch(1)
      if addContrastCheckbox:
        checkbox = QCheckBox('Improve contrast')
        widgets['contrastCheckbox'] = checkbox
        checkbox.toggled.connect(lambda: widgets['loop'].exit())
        QTimer.singleShot(0, lambda: checkbox.setChecked(hyperparameters["outputValidationVideoContrastImprovement"]))
        checkboxesLayout.addWidget(checkbox, alignment=Qt.AlignmentFlag.AlignCenter)
      if addUnprocessedFrameCheckbox:
        checkbox = QCheckBox('Show unprocessed frame')
        widgets['unprocessedFrameCheckbox'] = checkbox
        checkbox.toggled.connect(lambda: widgets['loop'].exit())
        checkboxesLayout.addWidget(checkbox, alignment=Qt.AlignmentFlag.AlignCenter)
      if addZoomCheckbox:
        numberOfAnimals = hyperparameters.get('nbAnimalsPerWell', 1)
        checkbox = QCheckBox('Zoom in to animal')
        checkbox.getAnimalIdx = lambda: animalSpinbox.value() if numberOfAnimals > 1 else 0
        widgets['zoomInCheckbox'] = checkbox
        checkbox.toggled.connect(lambda: widgets['loop'].exit())
        checkboxesLayout.addWidget(checkbox, alignment=Qt.AlignmentFlag.AlignCenter)
        if numberOfAnimals > 1:
          animalSpinbox = QSpinBox()
          animalSpinbox.setRange(0, numberOfAnimals - 1)
          animalSpinbox.valueChanged.connect(lambda: widgets['loop'].exit())
          checkboxesLayout.addWidget(animalSpinbox, alignment=Qt.AlignmentFlag.AlignLeft)
      checkboxesLayout.addStretch(1)
      layout.addLayout(checkboxesLayout, 2, 0, 1, 2, Qt.AlignmentFlag.AlignCenter)

    if l is not None:
      _createWidget(layout, status, widgets, (hyperparameters["firstFrame"], hyperparameters["lastFrame"] - 1, "You can also go through the video with the keys left arrow (backward); right arrow (forward); page down (fast backward); page up (fast forward)"), "Frame number", widgets, hasCheckbox)
    else:
      layout.addWidget(QWidget())
    for info, name in zip(organizationTab, hyperparametersListNames):
      _createWidget(layout, status, hyperparameters, info, name, widgets, hasCheckbox)

    mainLayout = QVBoxLayout()
    titleLabel = util.apply_style(QLabel(title), font=util.TITLE_FONT)
    titleLabel.setMinimumSize(1, 1)
    titleLabel.resizeEvent = lambda evt: titleLabel.setMinimumWidth(evt.size().width()) or titleLabel.setWordWrap(evt.size().width() <= titleLabel.sizeHint().width())
    mainLayout.addWidget(titleLabel, alignment=Qt.AlignmentFlag.AlignCenter)
    mainLayout.addLayout(layout)
    buttonsLayout = QHBoxLayout()
    buttonsLayout.addStretch()
    if documentationLink is not None:
      documentationBtn = util.apply_style(QPushButton("Help"), background_color="red")
      documentationBtn.clicked.connect(lambda: webbrowser.open_new(documentationLink))
      buttonsLayout.addWidget(documentationBtn, alignment=Qt.AlignmentFlag.AlignCenter)

    def saveClicked():
      widgets['saved'] = True
      widgets['loop'].exit()
    def discardClicked():
      widgets['discarded'] = True
      widgets['loop'].exit()
    for text, cb, color in (("Discard changes.", discardClicked, None), ("Done! Save changes!", saveClicked, util.DEFAULT_BUTTON_COLOR)):
      button = QPushButton(text) if color is None else util.apply_style(QPushButton(text), background_color=color)
      button.clicked.connect(cb)
      buttonsLayout.addWidget(button, alignment=Qt.AlignmentFlag.AlignCenter)
    buttonsLayout.addStretch()
    mainLayout.addLayout(buttonsLayout)
    temporaryPage = QWidget()
    temporaryPage.setLayout(mainLayout)
    stackedLayout.addWidget(temporaryPage)
    widgets['oldWidget'] = stackedLayout.currentWidget()
    stackedLayout.setCurrentWidget(temporaryPage)
    frame.setMinimumSize(1, 1)
    frame.show()
    util.setPixmapFromCv(frameToShow, frame)
  else:
    flattenedList = []
    for name in hyperparametersListNames:
      if isinstance(name, QAbstractListModel):
        flattenedList.extend((n for n in name.itemlist if n in hyperparameters))
      else:
        flattenedList.append(name)
    if 'steps' in hyperparameters:
      for idx in range(len(hyperparameters['steps'])):
        name = f'Step {idx + 1}'
        if name not in widgets:
          continue
        slider = widgets[name]
        value = hyperparameters['steps'][idx]
        if slider.value() != value:
          slider.setValue(value)
        minn = slider.minimum()
        maxx = slider.maximum()
        if (value - minn) > (maxx - minn) * 0.9:
          maxx = int(minn + (value - minn) * 1.1)
        elif maxx - minn > 255 and value < minn + (maxx - minn) * 0.1:
          maxx = int(minn + (maxx - minn) * 0.9)
        else:
          continue
        maxx = min(maxx, MAX_INT32)
        slider.setMaximum(maxx)
    for name in flattenedList:
      if name not in widgets:
        continue
      slider = widgets[name]
      value = len(hyperparameters[name]) if name == 'steps' else hyperparameters[name]
      if slider.value() != value:
        slider.setValue(value)
      if name.startswith("authorizedRelativeLengthTailEnd") or name.startswith("thetaDiffAccept") or name == "eyeBinaryThreshold":
        continue
      minn = slider.minimum()
      maxx = slider.maximum()
      if name == "frameGapComparision" and maxx == hyperparameters["lastFrame"] - hyperparameters["firstFrame"] - 1:
        continue
      if (value - minn) > (maxx - minn) * 0.9:
        maxx = int(minn + (value - minn) * 1.1)
      elif maxx - minn > 255 and value < minn + (maxx - minn) * 0.1:
        maxx = int(minn + (maxx - minn) * 0.9)
      else:
        continue
      if name == "frameGapComparision":
        maxx = min(maxx, hyperparameters["lastFrame"] - hyperparameters["firstFrame"] - 1)
      else:
        maxx = min(maxx, MAX_INT32)
      slider.setMaximum(maxx)
      localCursorX = slider.mapFromGlobal(QCursor.pos()).x()
      sliderLowerX = slider.sliderWidth() // 10
      sliderUpperX = slider.sliderWidth() - sliderLowerX
      if slider.isSliderDown() and (localCursorX <= sliderLowerX or localCursorX >= sliderUpperX):
        timer = QTimer()
        timers.append(timer)
        timer.setSingleShot(True)
        newPosition = int(min(localCursorX / slider.sliderWidth(), 1) * maxx)
        timer.timeout.connect(lambda slider=slider: (slider.setPosition(newPosition) if newPosition > 0 else widgets['loop'].exit()) or timers.remove(timer))
        timer.start(30)
    if 'frameSlider' in widgets:
      widgets['frameSlider'].setValue(l)
    widgets['frame'].clear()
    widgets['Frame number'] = l
    util.setPixmapFromCv(frameToShow, widgets['frame'])

  with app.suppressBusyCursor():
    widgets['loop'].exec()
  for timer in timers:
    timer.stop()

  if widgets['saved']:
    newParams = {}
    for name in hyperparametersListNames:
      if isinstance(name, QAbstractListModel):
        for name in name.itemlist:
          if name in hyperparameters:
            newParams[name] = hyperparameters[name]
      elif isinstance(name, tuple):
        for n in name:
          if hyperparameters.get(n, False):
            newParams[n] = hyperparameters[n]
      else:
        newParams[name] = hyperparameters[name]
    pickle.dump(newParams, open(os.path.join(paths.getRootDataFolder(), 'newhyperparameters'), 'wb'))
    temporaryPage = stackedLayout.currentWidget()
    stackedLayout.setCurrentWidget(widgets['oldWidget'])
    stackedLayout.removeWidget(temporaryPage)
    raise ValueError
  if widgets['discarded']:
    temporaryPage = stackedLayout.currentWidget()
    stackedLayout.setCurrentWidget(widgets['oldWidget'])
    stackedLayout.removeWidget(temporaryPage)
    raise NameError

  return widgets['Frame number'], widgets


def adjustDetectMouvRawVideosParams(img, res, l, totDiff, hyperparameters, widgets):
  documentationLink = "https://zebrazoom.org/documentation/docs/configurationFile/throughGUI/boutsDetection"
  hyperparametersListNames = ["frameGapComparision", "thresForDetectMovementWithRawVideo", "halfDiameterRoiBoutDetect", "minNbPixelForDetectMovementWithRawVideo"]
  organizationTab = [
  [0, 15,  "Increase if small movements are not detected. An increase that's too big could lead to a detection of bout that's too early however."],
  [0, 255, "Increase if too much movement is being detected."],
  [0, 500, "Controls the size of the images on which pixel change from image to the next are counted."],
  [0, 50,  "Increase if too much movement is being detected."],]
  title = "Red dot must appear only when movement is occuring"
  if widgets is not None and "contrastCheckbox" in widgets and widgets["contrastCheckbox"].isChecked():
    import zebrazoom.code.util as util
    img = util.improveContrast(img, hyperparameters["outputValidationVideoContrastImprovementQuartile"])
  frameToShow = np.concatenate((img, res),axis=1)
  frameToShow = cv2.cvtColor(frameToShow,cv2.COLOR_GRAY2RGB)

  minDimension    = min(len(frameToShow), len(frameToShow[0]))
  redDotDimension = 20
  if minDimension < 200:
    redDotDimension = int(minDimension / 10)

  fontSize = redDotDimension/20
  tickness = int(3*fontSize) if int(3*fontSize) != 0 else 1
  frameToShow = cv2.putText(frameToShow, str(l), (2*redDotDimension, redDotDimension), cv2.FONT_HERSHEY_SIMPLEX, fontSize, (0,255,0), tickness)

  if totDiff > hyperparameters["minNbPixelForDetectMovementWithRawVideo"]:
    cv2.circle(frameToShow, (redDotDimension, redDotDimension), redDotDimension, (0,0,255), -1)
  return adjustHyperparameters(l, hyperparameters, hyperparametersListNames, frameToShow, title, organizationTab, widgets, documentationLink=documentationLink, addContrastCheckbox=hyperparameters['headEmbeded'])


def adjustHeadEmbededTrackingParams(nbTailPoints, i, firstFrame, output, outputHeading, frame, frame2, hyperparameters, widgets):
  documentationLink = "https://zebrazoom.org/documentation/docs/configurationFile/throughGUI/headEmbeddedSwimming"
  # hyperparametersListNames = ["headEmbededAutoSet_BackgroundExtractionOption", "overwriteFirstStepValue", "overwriteLastStepValue", "overwriteNbOfStepValues", "headEmbededParamTailDescentPixThreshStopOverwrite", "authorizedRelativeLengthTailEnd"]
  hyperparametersListNames = ["headEmbededAutoSet_BackgroundExtractionOption", "overwriteFirstStepValue", "overwriteLastStepValue", "headEmbededParamTailDescentPixThreshStopOverwrite", "authorizedRelativeLengthTailEnd", "overwriteHeadEmbededParamGaussianBlur"]
  organizationTab = [
  [0, 20, "Transforms non-background pixels to black. Can be useful when the tail isn't very different from the background."],
  [0, 80, "Increase this to avoid having the tracking points go on the head instead of the tail."],
  [1, 80, 'Increase this if the tail tracking is getting off track "mid-tail". Decrease if the tail tracking is going too far (further than the tip).'],
  # [1, 50, "This is set automatically when you change either overwriteFirstStepValue or overwriteLastStepValue. Decrease to make the tracking faster."],
  [0, 255, "ALMOST ALWAYS IGNORE THIS PARAMETER. Decrease if the tail tracking is going too far (further than the tip of the tail). Increase if the tail if not going far enough (stops before the tip)."],
  [0, 1, 'ALMOST ALWAYS IGNORE THIS PARAMETER. Relative length along the "normal lenght" of the tail where the tracking is "allowed" to stop. Decrease if the tail becomes invisible "mid-tail".'],
  [0, 100, 'THIS PARAMETER CAN USUALLY BE IGNORED. Try to find the right balance between too much and too little gaussian smoothing of the image.'],]
  title = "Tracking"

  # frame2 = np.concatenate((frame2, frame),axis=1)

  if widgets is not None and widgets["contrastCheckbox"].isChecked():
    import zebrazoom.code.util as util
    frame2 = util.improveContrast(frame2, hyperparameters["outputValidationVideoContrastImprovementQuartile"])

  frame2 = cv2.cvtColor(frame2,cv2.COLOR_GRAY2RGB)
  for k in range(0, hyperparameters["nbAnimalsPerWell"]):
    for j in range(0, nbTailPoints):
      x = int(output[k, i-firstFrame][j][0])
      y = int(output[k, i-firstFrame][j][1])
      cv2.circle(frame2, (x, y), 1, (0, 255, 0),   -1)
    x = int(output[k, i-firstFrame][nbTailPoints-1][0])
    y = int(output[k, i-firstFrame][nbTailPoints-1][1])
    cv2.circle(frame2, (x, y), 2, (0, 0, 255),   -1)
    x = output[k, i-firstFrame][0][0]
    y = output[k, i-firstFrame][0][1]

  return adjustHyperparameters(i, hyperparameters, hyperparametersListNames, frame2, title, organizationTab, widgets, documentationLink=documentationLink, addContrastCheckbox=True)


def adjustFreelySwimTrackingParams(nbTailPoints, i, firstFrame, output, outputHeading, frame, frame2, hyperparameters, widgets):
  if hyperparameters["trackTail"] == 1:
    hyperparametersListNames = ["minPixelDiffForBackExtract", "maxAreaBody", "minTailSize", "maxTailSize"]
    organizationTab = [
    [0, 20, "Increase this if some of the background is not completely white. Decrease if you can't see all of the animals. "],
    [0, 20, "Try increasing this if no tracking is showing."],
    [0, 20, "Try increasing this if no tracking is showing."],
    [0, 20, "Try decreasing this if no tracking is showing."],]
  else:
    hyperparametersListNames = ["minPixelDiffForBackExtract"]
    # The gaussian image filter should be added below, and maybe also the trajectories post-processing option
    organizationTab = [
    [0, 20, "Increase this if some of the background is not completely white. Decrease if you can't see all of the animals. "],]

  title = "Adjust parameters in order for the background to be white and the animals to be gray/black."

  frame2 = cv2.cvtColor(frame2,cv2.COLOR_GRAY2RGB)
  for k in range(0, hyperparameters["nbAnimalsPerWell"]):
    if hyperparameters["trackTail"] == 1:
      for j in range(0, nbTailPoints):
        x = int(output[k, i-firstFrame][j][0])
        y = int(output[k, i-firstFrame][j][1])
        cv2.circle(frame2, (x, y), 1, (0, 255, 0),   -1)
    x = int(output[k, i-firstFrame][nbTailPoints-1][0])
    y = int(output[k, i-firstFrame][nbTailPoints-1][1])
    cv2.circle(frame2, (x, y), 2, (0, 0, 255),   -1)
    x = output[k, i-firstFrame][0][0]
    y = output[k, i-firstFrame][0][1]

  return adjustHyperparameters(i, hyperparameters, hyperparametersListNames, frame2, title, organizationTab, widgets)


def adjustFreelySwimTrackingAutoParams(nbTailPoints, i, firstFrame, output, outputHeading, frame, frame2, hyperparameters, widgets):
  documentationLink = "https://zebrazoom.org/documentation/docs/configurationFile/throughGUI/trackingFreelySwimming"
  hyperparametersListNames = ["recalculateForegroundImageBasedOnBodyArea", "adjustMinPixelDiffForBackExtract_nbBlackPixelsMax", "minPixelDiffForBackExtract"]
  organizationTab = [
  [0, 1, "Set to 1 for Method 3 and to 0 for Method 1 or 2"],
  [0, 20, "Set to 0 for Method 1. For method 2 and 3: increase if the tip of tail is detected to soon, decrease if the tracking looks messy."],
  [0, 20, "For method 1: decrease if the tip of tail is detected to soon, increase if the tracking looks messy"],]

  title = "Adjust parameters in order for the background to be white and the animals to be gray/black."

  frame2 = cv2.cvtColor(frame2,cv2.COLOR_GRAY2RGB)
  for k in range(0, hyperparameters["nbAnimalsPerWell"]):
    if hyperparameters["trackTail"] == 1:
      for j in range(0, nbTailPoints):
        x = int(output[k, i-firstFrame][j][0])
        y = int(output[k, i-firstFrame][j][1])
        cv2.circle(frame2, (x, y), 1, (0, 255, 0),   -1)
    x = int(output[k, i-firstFrame][nbTailPoints-1][0])
    y = int(output[k, i-firstFrame][nbTailPoints-1][1])
    cv2.circle(frame2, (x, y), 2, (0, 0, 255),   -1)
    x = output[k, i-firstFrame][0][0]
    y = output[k, i-firstFrame][0][1]

  return adjustHyperparameters(i, hyperparameters, hyperparametersListNames, frame2, title, organizationTab, widgets, documentationLink=documentationLink)


def adjustHeadEmbeddedEyeTrackingParamsSegment(i, frame, hyperparameters, widgets):
  documentationLink = None
  hyperparametersListNames = ["eyeTrackingHeadEmbeddedHalfDiameter", "eyeTrackingHeadEmbeddedWidth"]
  organizationTab = [
  [1, 255, ""],
  [1, 255, ""],]
  title = "Eye Tracking"

  return adjustHyperparameters(i, hyperparameters, hyperparametersListNames, frame, title, organizationTab, widgets, documentationLink=documentationLink)


def adjustHeadEmbeddedEyeTrackingParamsEllipse(i, frame, hyperparameters, widgets):
  documentationLink = None
  hyperparametersListNames = ["eyeBinaryThreshold", "eyeFilterKernelSize", "eyeTrackingHeadEmbeddedHalfDiameter"]
  organizationTab = [
  [0, 255, ""],
  [1, 255, ""],
  [1, 255, ""],]
  title = "Eye Tracking"

  return adjustHyperparameters(i, hyperparameters, hyperparametersListNames, frame, title, organizationTab, widgets, documentationLink=documentationLink)



import json
import os

import cv2
import numpy as np

from PyQt5.QtWidgets import QApplication

import zebrazoom.code.util as util
from zebrazoom.code.adjustHyperparameters import adjustFreelySwimTrackingParams, adjustFreelySwimTrackingAutoParams, adjustHeadEmbededTrackingParams
from zebrazoom.code.tracking.tracking import Tracking
from zebrazoom.code.tracking import register_tracking_method
from ._base import BaseGUITrackingMethod


class GUITracking(Tracking, BaseGUITrackingMethod):
  def _adjustParameters(self, i, initialCurFrame, frame, frame2, back, widgets):
    if self._hyperparameters["adjustHeadEmbededTracking"] == 1:
      i, widgets = adjustHeadEmbededTrackingParams(self._nbTailPoints, i, self._firstFrame, self._trackingHeadTailAllAnimals, self._trackingHeadingAllAnimals, frame, frame2, self._hyperparameters, widgets)
      self._adjustHeadEmbededHyperparameters(frame)
    elif self._hyperparameters["adjustFreelySwimTracking"] == 1:
      i, widgets = adjustFreelySwimTrackingParams(self._nbTailPoints, i, self._firstFrame, self._trackingHeadTailAllAnimals, self._trackingHeadingAllAnimals, frame, frame2, self._hyperparameters, widgets)
    elif self._hyperparameters["adjustFreelySwimTrackingAutomaticParameters"] == 1:
      # Preparing image to show
      if self._hyperparameters["recalculateForegroundImageBasedOnBodyArea"] == 1:
        minPixelDiffForBackExtract = self._hyperparameters["minPixelDiffForBackExtractBody"]
      else:
        if self._hyperparameters["adjustMinPixelDiffForBackExtract_nbBlackPixelsMax"]:
          minPixelDiffForBackExtract = self._hyperparameters["minPixelDiffForBackExtractHead"]
          del self._hyperparameters["minPixelDiffForBackExtractHead"] # Not sure why this is necessary: need to check the code to make sure there isn't a bug somewhere
        else:
          minPixelDiffForBackExtract = self._hyperparameters["minPixelDiffForBackExtract"]
      curFrame = initialCurFrame
      putToWhite = (curFrame.astype('int32') >= (back.astype('int32') - minPixelDiffForBackExtract) )
      curFrame[putToWhite] = 255
      ret, frame2 = cv2.threshold(curFrame, self._hyperparameters["thresholdForBlobImg"], 255, cv2.THRESH_BINARY)
      # Showing current image and waiting for next parameter/frame change
      i, widgets = adjustFreelySwimTrackingAutoParams(self._nbTailPoints, i, self._firstFrame, self._trackingHeadTailAllAnimals, self._trackingHeadingAllAnimals, frame, frame2, self._hyperparameters, widgets)
      # Puts self._hyperparameters values to accepted values
      self._hyperparameters["recalculateForegroundImageBasedOnBodyArea"] = 0 if self._hyperparameters["recalculateForegroundImageBasedOnBodyArea"] < 0.5 else 1
      if self._hyperparameters["adjustMinPixelDiffForBackExtract_nbBlackPixelsMax"] < 0:
        self._hyperparameters["adjustMinPixelDiffForBackExtract_nbBlackPixelsMax"] = 0
      if self._hyperparameters["minPixelDiffForBackExtract"] < 0:
        self._hyperparameters["minPixelDiffForBackExtract"] = 0
      else:
        if self._hyperparameters["minPixelDiffForBackExtract"] > 255:
          self._hyperparameters["minPixelDiffForBackExtract"] = 255
      self._hyperparameters["minPixelDiffForBackExtract"] = int(self._hyperparameters["minPixelDiffForBackExtract"])
    else:
      return None
    return i, widgets

  def _getCoordinates(self, frame, title, zoomable, dialog):
    return util.getPoint(frame, title, zoomable=zoomable, dialog=dialog)

  def _addBlackLineToImgSetParameters(self, frame):
    hyperparametersToSave = {"addBlackLineToImg_Width": 0}

    frame2 = frame.copy()
    # frame2 = 255 - frame2
    quartileChose = self._hyperparameters["outputValidationVideoContrastImprovementQuartile"]
    lowVal  = int(np.quantile(frame2, quartileChose))
    highVal = int(np.quantile(frame2, 1 - quartileChose))
    frame2[frame2 < lowVal]  = lowVal
    frame2[frame2 > highVal] = highVal
    frame2 = frame2 - lowVal
    mult  = np.max(frame2)
    frame2 = frame2 * (255/mult)
    frame2 = frame2.astype('uint8')

    addNewSegment  = True

    while addNewSegment:

      startPoint = self._getCoordinates(frame2, "Click on beginning of segment to set to black pixels", False, True)
      endPoint = util.getLine(frame2, "Click on the end of the segment to set to black pixels", self._hyperparameters["addBlackLineToImg_Width"], startPoint)

      if not("imagePreProcessMethod" in self._hyperparameters) or type(self._hyperparameters["imagePreProcessMethod"]) == int or len(self._hyperparameters["imagePreProcessMethod"]) == 0:
        self._hyperparameters["imagePreProcessMethod"]     = ["setImageLineToBlack"]
        self._hyperparameters["imagePreProcessParameters"] = [[startPoint[0], startPoint[1], endPoint[0], endPoint[1], self._hyperparameters["addBlackLineToImg_Width"]]]
        hyperparametersToSave["imagePreProcessMethod"]     = ["setImageLineToBlack"]
        hyperparametersToSave["imagePreProcessParameters"] = [[startPoint[0], startPoint[1], endPoint[0], endPoint[1], self._hyperparameters["addBlackLineToImg_Width"]]]
      else:
        self._hyperparameters["imagePreProcessMethod"] = self._hyperparameters["imagePreProcessMethod"] + ["setImageLineToBlack"]
        self._hyperparameters["imagePreProcessParameters"] += [[startPoint[0], startPoint[1], endPoint[0], endPoint[1], self._hyperparameters["addBlackLineToImg_Width"]]]
        hyperparametersToSave["imagePreProcessMethod"]     += ["setImageLineToBlack"]
        hyperparametersToSave["imagePreProcessParameters"] += [[startPoint[0], startPoint[1], endPoint[0], endPoint[1], self._hyperparameters["addBlackLineToImg_Width"]]]

      print('addBlackLineToImg_Width is at', self._hyperparameters["addBlackLineToImg_Width"],', so in the configuration file, "imagePreProcessMethod" has just been set to ["setImageLineToBlack"] and "imagePreProcessParameters" has just been set to', str(self._hyperparameters["imagePreProcessParameters"]))

      frame2 = cv2.line(frame2, (startPoint[0], startPoint[1]), (endPoint[0], endPoint[1]), (255, 255, 255), self._hyperparameters["addBlackLineToImg_Width"])

      def doneAddingSegments():
        nonlocal addNewSegment
        addNewSegment = False
      buttons = (("I want to add another segment.", None), ("I've added enough segments.", doneAddingSegments))
      self._debugFrame(frame2, title="Do you want to add another segment?", buttons=buttons)

  def findHeadPositionByUserInput(self, frame, frameNumber, wellNumber):
    plus = 0

    def tailNotStraight(frameWidget):
      nonlocal plus
      plus += 1
      util.setPixmapFromCv(self.headEmbededFrame(frameNumber + plus, wellNumber)[0], frameWidget, zoomable=True)
    return list(util.getPoint(np.uint8(frame * 255), "Click on the base of the tail", zoomable=True, extraButtons=(("Tail is not straight", tailNotStraight, False),),
                              dialog=not hasattr(QApplication.instance(), 'window')))

  def findTailTipByUserInput(self, frame, frameNumber, wellNumber):
    plus = 0

    def tailNotStraight(frameWidget):
      nonlocal plus
      plus += 1
      util.setPixmapFromCv(self.headEmbededFrame(frameNumber + plus, wellNumber)[0], frameWidget, zoomable=True)
    return list(util.getPoint(np.uint8(frame * 255), "Click on tail tip", zoomable=True, extraButtons=(("Tail is not straight", tailNotStraight, False),),
                              dialog=not hasattr(QApplication.instance(), 'window')))


register_tracking_method('tracking', GUITracking)



from zebrazoom.code.createValidationVideo import calculateInfoFrameForFrame, processFrame
from zebrazoom.code.getHyperparameters import getHyperparametersSimple

from ._createSuperStructFromH5 import createSuperStructFromH5
from ._openResultsFile import openResultsFile


def plotSingleFrameTrackingPoints(videoName: str):
  with openResultsFile(videoName, 'r') as results:
    hyperparameters = getHyperparametersSimple(dict(results['configurationFileUsed'].attrs))
    superStruct = createSuperStructFromH5(results)
    colorModifTab = [{"red": 0, "green": 0, "blue": 0}]
    colorModifTab.extend([{"red": random.randrange(255), "green": random.randrange(255), "blue": random.randrange(255)} for i in range(1, hyperparameters["nbAnimalsPerWell"])])
    infoFrame = []
    for i in range(len(superStruct["wellPoissMouv"])):
      for j in range(len(superStruct["wellPoissMouv"][i])):
        # assume we only have one bout
        bout = superStruct["wellPoissMouv"][i][j][0]
        infoFrame.extend(calculateInfoFrameForFrame(superStruct, hyperparameters, i, j, 0, bout["BoutStart"], colorModifTab))
    exampleFrame = results['exampleFrame'][:]
    if hyperparameters['invertBlackWhiteOnImages']:
      exampleFrame = 255 - exampleFrame
    return processFrame(exampleFrame, hyperparameters, infoFrame, colorModifTab)



import zebrazoom.code.tracking
from zebrazoom.code.tracking.customTrackingImplementations.examples.exampleTrackingMethod import ExampleTrackingMethod


class ExampleGUITrackingMethod(ExampleTrackingMethod):
  pass


zebrazoom.code.tracking.register_tracking_method('zebrazoom.example', ExampleGUITrackingMethod)



import os

import cv2
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np

from .plotSingleFrameTrackingPoints import plotSingleFrameTrackingPoints
from .getCurvaturePerBout import getCurvaturePerBout
from .getFPSandPixelSize import getFPSandPixelSize
from .plotCurvatureYaxisApproximate import plotCurvatureYaxisApproximateImplementation
from .getDataPerBout import getDataPerBout

def getSingleFrameTrackingAndCurvatureForFolder(pathToFolder: str):
  for basename in os.listdir(pathToFolder):
    if not basename.endswith('.h5'):
      continue
    filename = os.path.join(pathToFolder, basename)
    outputFolder = os.path.join(pathToFolder, 'output')
    if not os.path.exists(outputFolder):
      os.mkdir(outputFolder)
    _, pixelSize = getFPSandPixelSize(filename)
    [curvature, xTimeValues, yDistanceAlongTheTail] = getCurvaturePerBout(filename, 0, 0, 0)
    plotCurvatureYaxisApproximateImplementation(*[curvature, xTimeValues, yDistanceAlongTheTail], 1, pixelSize)
    plt.savefig(os.path.join(outputFolder, basename.replace('.h5', '_curvature.png')), dpi=300, bbox_inches='tight')
    plt.close()
    cv2.imwrite(os.path.join(outputFolder, basename.replace('.h5', '.png')), plotSingleFrameTrackingPoints(filename))
    
    TailPosX = getDataPerBout(filename, 0, 0, 0, 'TailPosX')
    TailPosY = getDataPerBout(filename, 0, 0, 0, 'TailPosY')
    
    pd.DataFrame(np.column_stack((np.insert(yDistanceAlongTheTail[:, 0][::-1], 0, 0), TailPosX[0, :], TailPosY[0, :], np.insert(curvature[:, 0][::-1], 0, 0))), columns=['distanceAlongTheTail', 'TailPosX', 'TailPosY', 'curvature']).to_excel(os.path.join(outputFolder, basename.replace('.h5', '.xlsx')))
    


import math
import os

import cv2
import numpy as np
from numpy import linspace
from scipy.interpolate import interp1d
from scipy.interpolate import splprep, splev
from scipy.optimize import curve_fit

from ._centerOfMassTailTracking import CenterOfMassTailTrackingMixin
from ._tailExtremityTracking import TailTrackingExtremityDetectMixin
from ._headEmbeddedTailTracking import HeadEmbeddedTailTrackingMixin
from ._headEmbeddedTailTrackingTeresaNicolson import HeadEmbeddedTailTrackingTeresaNicolsonMixin
from ._tailTrackingBlobDescent import TailTrackingBlobDescentMixin


class TailTrackingMixin(HeadEmbeddedTailTrackingMixin, CenterOfMassTailTrackingMixin, HeadEmbeddedTailTrackingTeresaNicolsonMixin, TailTrackingBlobDescentMixin, TailTrackingExtremityDetectMixin):
  def _tailTracking(self, animalId, i, frame, thresh1, threshForBlackFrames, thetaDiffAccept, trackingHeadTailAllAnimals, trackingHeadingAllAnimals, lastFirstTheta, maxDepth, tailTip, initialCurFrame, back, wellNumber=-1, xmin=0, ymin=0):
    headPosition = [trackingHeadTailAllAnimals[animalId, i-self._firstFrame][0][0]-xmin, trackingHeadTailAllAnimals[animalId, i-self._firstFrame][0][1]-ymin]
    heading      = trackingHeadingAllAnimals[animalId, i-self._firstFrame]

    if (self._hyperparameters["headEmbeded"] == 1):
      # through the "head embeded" method, either through "segment descent" or "center of mass descent"

      if self._hyperparameters["headEmbededTeresaNicolson"] == 1:
        oppHeading = (heading + math.pi) % (2 * math.pi)
        trackingHeadTailAllAnimalsI = self._headEmbededTailTrackingTeresaNicolson(headPosition, frame, maxDepth, tailTip, threshForBlackFrames)
      else:
        oppHeading = (heading + math.pi) % (2 * math.pi) # INSERTED FOR THE REFACTORING
        if self._hyperparameters["centerOfMassTailTracking"] == 0:
          trackingHeadTailAllAnimalsI = self._headEmbededTailTracking(headPosition, i, frame, maxDepth, tailTip)
        else:
          trackingHeadTailAllAnimalsI = self._centerOfMassTailTracking(headPosition, frame, maxDepth)

      if len(trackingHeadTailAllAnimalsI[0]) == len(trackingHeadTailAllAnimals[animalId, i-self._firstFrame]):
        trackingHeadTailAllAnimals[animalId, i-self._firstFrame] = trackingHeadTailAllAnimalsI

    else:
      if self._hyperparameters["freeSwimmingTailTrackingMethod"] == "tailExtremityDetect":
        # through the tail extremity descent method (original C++ method)
        [trackingHeadTailAllAnimalsI, newHeading] = self._tailTrackingExtremityDetect(headPosition, i, thresh1, frame, self._hyperparameters["debugTrackingPtExtreme"], heading, initialCurFrame, back, wellNumber)
        trackingHeadTailAllAnimals[animalId, i-self._firstFrame] = trackingHeadTailAllAnimalsI
        if newHeading != -1:
          trackingHeadingAllAnimals[animalId, i-self._firstFrame] = newHeading
      elif self._hyperparameters["freeSwimmingTailTrackingMethod"] == "blobDescent":
        # through the "blob descent" method
        trackingHeadTailAllAnimals[animalId, i-self._firstFrame] = self._tailTrackingBlobDescent(headPosition, i, thresh1, frame, lastFirstTheta, self._hyperparameters["debugTrackingPtExtreme"], thetaDiffAccept)
      else: # self._hyperparameters["freeSwimmingTailTrackingMethod"] == "none"
        # only tracking the head, not the tail
        trackingHeadTailAllAnimals[animalId, i-self._firstFrame][0][0] = headPosition[0]
        trackingHeadTailAllAnimals[animalId, i-self._firstFrame][0][1] = headPosition[1]



import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import json
import os

def alternativeKinematicParameterCalculation(args):

  if os.path.exists(os.path.join('zebrazoom/dataAnalysis/experimentOrganizationExcel', args.nameOfExperiment + '.xls')):
    excelOrganizationFilePath = os.path.join('zebrazoom/dataAnalysis/experimentOrganizationExcel', args.nameOfExperiment + '.xls')
  elif os.path.exists(os.path.join('zebrazoom/dataAnalysis/experimentOrganizationExcel', args.nameOfExperiment + '.xlsx')):
    excelOrganizationFilePath = os.path.join('zebrazoom/dataAnalysis/experimentOrganizationExcel', args.nameOfExperiment + '.xlsx')
  else:
    print("Excel organization file not found, is its extension .xls or .xlsx?")
    return
  
  excelOrganizationFile = pd.read_excel(excelOrganizationFilePath)

  allGenotypes  = []
  allConditions = []
  for videoId in range(len(excelOrganizationFile)):
    path      = excelOrganizationFile.loc[videoId]['path']
    trial_id  = excelOrganizationFile.loc[videoId]['trial_id']
    fq        = excelOrganizationFile.loc[videoId]['fq']
    pixelsize = excelOrganizationFile.loc[videoId]['pixelsize']
    genotype  = [gen.replace("'", "").replace('"', "").replace(' ', "") for gen in excelOrganizationFile.loc[videoId]['genotype'][1:-1].split(",")]
    condition = [cond.replace("'", "").replace('"', "").replace(' ', "") for cond in excelOrganizationFile.loc[videoId]['condition'][1:-1].split(",")]
    include   = [int(incl) for incl in excelOrganizationFile.loc[videoId]['include'][1:-1].split(",")]
    allGenotypes.append(genotype)
    allConditions.append(condition)
    if len(include) != len(condition) or len(include) != len(genotype) or len(genotype) != len(condition):
      print("inconsistent include, condition, genotype!!!")
      print(len(include), len(condition), len(genotype))
      return
  allGenotypes  = np.unique(allGenotypes).tolist()
  allConditions = np.unique(allConditions).tolist()

  GenXCondDictAllBouts  = {}
  GenXCondDictPerAnimal = {}
  for gen in allGenotypes:
    GenXCondDictAllBouts[gen] = {}
    GenXCondDictPerAnimal[gen] = {}
    for cond in allConditions:
      GenXCondDictAllBouts[gen][cond] = []
      GenXCondDictPerAnimal[gen][cond] = []
  
  for videoId in range(len(excelOrganizationFile)):
    path      = excelOrganizationFile.loc[videoId]['path']
    trial_id  = excelOrganizationFile.loc[videoId]['trial_id']
    fq        = excelOrganizationFile.loc[videoId]['fq']
    pixelsize = excelOrganizationFile.loc[videoId]['pixelsize']
    genotype  = [gen.replace("'", "").replace('"', "").replace(' ', "") for gen in excelOrganizationFile.loc[videoId]['genotype'][1:-1].split(",")]
    condition = [cond.replace("'", "").replace('"', "").replace(' ', "") for cond in excelOrganizationFile.loc[videoId]['condition'][1:-1].split(",")]
    include   = [int(incl) for incl in excelOrganizationFile.loc[videoId]['include'][1:-1].split(",")]
    pathToResultFile = os.path.join(os.path.join('zebrazoom/ZZoutput', trial_id), 'results_' + trial_id + '.txt') if path == "defaultZZoutputFolder" else os.path.join(path, 'results_' + trial_id + '.txt')
    with open(pathToResultFile) as f:
      supstruct = json.load(f)
      for numWell in range(0, len(supstruct['wellPoissMouv'])):
        if include[numWell]:
          for numAnimal in range(0, len(supstruct['wellPoissMouv'][numWell])):
            perAnimal = []
            for numBout in range(0, len(supstruct['wellPoissMouv'][numWell][numAnimal])):
              gen  = genotype[numWell]
              cond = condition[numWell]
              bout = supstruct['wellPoissMouv'][numWell][numAnimal][numBout]
              BoutStart  = bout['BoutStart']
              BoutEnd    = bout['BoutEnd']
              BoutLength = (BoutEnd - BoutStart + 1) / fq
              GenXCondDictAllBouts[gen][cond].append(BoutLength)
              perAnimal.append(BoutLength)
            GenXCondDictPerAnimal[gen][cond].append(perAnimal)
  
  if 'WT' in allGenotypes:
    allGenotypes.remove('WT')
    allGenotypes.insert(0, 'WT')
  
  # All bouts combined plot
  
  allLabels = []
  allData   = [] 
  for cond in allConditions:
    for gen in allGenotypes:
      allLabels.append(gen + '_' + cond)
      allData.append(GenXCondDictAllBouts[gen][cond])
  
  fig, ax = plt.subplots()
  bp = ax.boxplot(allData)
  ax.set_xticklabels(allLabels)
  ax.set_ylabel('Duration')
  ax.set_title('All bouts combined')
  plt.show()
  
  # Median per fish plot
  
  allLabels = []
  allData   = [] 
  for cond in allConditions:
    for gen in allGenotypes:
      allLabels.append(gen + '_' + cond)
      allData.append([np.median(arr) for arr in GenXCondDictPerAnimal[gen][cond]])
  
  fig, ax = plt.subplots()
  bp = ax.boxplot(allData)
  ax.set_xticklabels(allLabels)
  ax.set_ylabel('Duration')
  ax.set_title('Median per fish')
  plt.show()  



import cv2
import numpy as np

import zebrazoom.code.util as util
from zebrazoom.code.adjustHyperparameters import adjustHeadEmbeddedEyeTrackingParamsEllipse, adjustHeadEmbeddedEyeTrackingParamsSegment
from zebrazoom.code.tracking import BaseZebraZoomTrackingMethod


class BaseGUITrackingMethod(BaseZebraZoomTrackingMethod):
  def _debugFrame(self, frame, title=None, buttons=(), timeout=None):
    util.showFrame(frame, title=title, buttons=buttons, timeout=timeout)

  def _debugTracking(self, frameNumber: int, output: list, outputHeading: list, frame2: np.array) -> None:
    if not self._hyperparameters["debugTracking"]:
      return

    if type(frame2[0][0]) == int or type(frame2[0][0]) == np.uint8:
      frame2 = cv2.cvtColor(frame2,cv2.COLOR_GRAY2RGB)

    for k in range(0, self._hyperparameters["nbAnimalsPerWell"]):
      for j in range(0, self._nbTailPoints):
        x = int(output[k, frameNumber-self._firstFrame][j][0])
        y = int(output[k, frameNumber-self._firstFrame][j][1])
        cv2.circle(frame2, (x, y), 1, (0, 255, 0),   -1)
      x = int(output[k, frameNumber-self._firstFrame][self._nbTailPoints-1][0]) # it used to be 10 instead of 9 here, not sure why
      y = int(output[k, frameNumber-self._firstFrame][self._nbTailPoints-1][1]) # it used to be 10 instead of 9 here, not sure why
      cv2.circle(frame2, (x, y), 2, (0, 0, 255),   -1)

      x = output[k, frameNumber-self._firstFrame][0][0]
      y = output[k, frameNumber-self._firstFrame][0][1]
      # cv2.line(frame2, (int(x / getRealValueCoefX), int(y / getRealValueCoefY)), (int((x+20*math.cos(outputHeading[k, i-self._firstFrame])) / getRealValueCoefX), int((y+20*math.sin(outputHeading[k, i-self._firstFrame])) / getRealValueCoefY)), (255,0,0), 3)

    if self._hyperparameters["debugTrackingPtExtremeLargeVerticals"]: # Put this to True for large resolution videos (to be able to see on your screen what's happening)
      frame2 = frame2[int(y-200):len(frame2), :]

    self._debugFrame(frame2, title='Tracked frame')

  def _adjustHeadEmbeddedEyeTrackingParamsSegment(self, i, colorFrame, widgets):
    return adjustHeadEmbeddedEyeTrackingParamsSegment(i, colorFrame, self._hyperparameters, widgets)

  def _adjustHeadEmbeddedEyeTrackingParamsEllipse(self, i, colorFrame, widgets):
    return adjustHeadEmbeddedEyeTrackingParamsEllipse(i, colorFrame, self._hyperparameters, widgets)



import cv2

def removeAnimalJustTrackedFromFrame(self, frame, wellNumber, animalId, k):
  if "largerPixelRemoval" in self._hyperparameters and self._hyperparameters["largerPixelRemoval"]:
    frame = cv2.circle(frame.copy(), (int(self._trackingDataPerWell[wellNumber][animalId][k][0][0]), int(self._trackingDataPerWell[wellNumber][animalId][k][0][1])), int(self._hyperparameters["maxDepth"]/3), (255, 255, 255), -1)
    for pointOnTail in range(1, len(self._trackingDataPerWell[wellNumber][animalId][k])):
      start_point = (int(self._trackingDataPerWell[wellNumber][animalId][k][pointOnTail-1][0]), int(self._trackingDataPerWell[wellNumber][animalId][k][pointOnTail-1][1]))
      end_point   = (int(self._trackingDataPerWell[wellNumber][animalId][k][pointOnTail][0]), int(self._trackingDataPerWell[wellNumber][animalId][k][pointOnTail][1]))
      if pointOnTail == 1:
        end_point   = (int(3 * self._trackingDataPerWell[wellNumber][animalId][k][pointOnTail-1][0] - 2 * self._trackingDataPerWell[wellNumber][animalId][k][pointOnTail][0]), int(3 * self._trackingDataPerWell[wellNumber][animalId][k][pointOnTail-1][1] - 2 * self._trackingDataPerWell[wellNumber][animalId][k][pointOnTail][1]))
      if pointOnTail == len(self._trackingDataPerWell[wellNumber][animalId][k]) - 1:
        end_point   = (int(2 * self._trackingDataPerWell[wellNumber][animalId][k][pointOnTail][0] - self._trackingDataPerWell[wellNumber][animalId][k][pointOnTail-1][0]), int(2 * self._trackingDataPerWell[wellNumber][animalId][k][pointOnTail][1] - self._trackingDataPerWell[wellNumber][animalId][k][pointOnTail-1][1]))
      cv2.line(frame, start_point, end_point, (255, 255, 255), int(self._hyperparameters["maxDepth"]/3))
  else:
    frame = cv2.circle(frame.copy(), (int(self._wellPositions[wellNumber]['topLeftX'] + self._trackingDataPerWell[wellNumber][animalId][k][0][0]), int(self._wellPositions[wellNumber]['topLeftY'] + self._trackingDataPerWell[wellNumber][animalId][k][0][1])), int(self._hyperparameters["maxDepth"]/4), (255, 255, 255), -1)
    for pointOnTail in range(1, len(self._trackingDataPerWell[wellNumber][animalId][k])):
      start_point = (int(self._wellPositions[wellNumber]['topLeftX'] + self._trackingDataPerWell[wellNumber][animalId][k][pointOnTail-1][0]), int(self._wellPositions[wellNumber]['topLeftY'] + self._trackingDataPerWell[wellNumber][animalId][k][pointOnTail-1][1]))
      end_point   = (int(self._wellPositions[wellNumber]['topLeftX'] + self._trackingDataPerWell[wellNumber][animalId][k][pointOnTail][0]), int(self._wellPositions[wellNumber]['topLeftY'] + self._trackingDataPerWell[wellNumber][animalId][k][pointOnTail][1]))
      cv2.line(frame, start_point, end_point, (255, 255, 255), int(self._hyperparameters["maxDepth"]/5))
  
  return frame



import os
import sys
import numpy as np

from PyQt5.QtCore import pyqtSignal, QTimer
from PyQt5.QtWidgets import QPlainTextEdit, QVBoxLayout

import zebrazoom.code.paths as paths
import zebrazoom.code.util as util
from zebrazoom.code.vars import getGlobalVariables
globalVariables = getGlobalVariables()


class _PopUpAlgoFollowPlainTextEdit(QPlainTextEdit):
  finished = pyqtSignal()


def _update(textedit, timer):
  if globalVariables["mac"] != 1 and globalVariables["lin"] != 1:
    with open(os.path.join(paths.getRootDataFolder(), "trace.txt"), "r") as f:
      contents = f.read()
    textedit.setPlainText(contents)
    if "ZebraZoom Analysis finished for" in contents:
      textedit.finished.emit()


def createTraceFile(msg):
  if globalVariables["mac"] != 1 and globalVariables["lin"] != 1:
    with open(os.path.join(paths.getRootDataFolder(), "trace.txt"), "w+") as f:
      f.write(msg)


def initialise():
  if globalVariables["mac"] != 1 and globalVariables["lin"] != 1:
    from zebrazoom.GUIAllPy import PlainApplication
    app = PlainApplication(sys.argv)
    layout = QVBoxLayout()
    textedit = _PopUpAlgoFollowPlainTextEdit()
    textedit.setFixedSize(600, 400)
    textedit.setReadOnly(True)
    layout.addWidget(textedit)
    timer = QTimer()
    timer.setInterval(1)
    timer.timeout.connect(lambda: _update(textedit, timer))
    timer.start()
    util.showDialog(layout, title="Tracking in Progress", exitSignals=(textedit.finished,))
    timer.stop()


def prepend(text):
  if globalVariables["mac"] != 1 and globalVariables["lin"] != 1:
    with open(os.path.join(paths.getRootDataFolder(), "trace.txt"), "r+") as f:
      content = f.read()
      f.seek(0, 0)
      f.write(text.rstrip('\r\n') + '\n' + content)



import numpy as np


def createSuperStructFromH5(results):
  superStruct = {}
  superStruct["firstFrame"] = results.attrs['firstFrame']
  superStruct['lastFrame'] = results.attrs['lastFrame']
  if 'videoFPS' in results.attrs:
    superStruct['videoFPS'] = results.attrs['videoFPS']
  if 'videoPixelSize' in results.attrs:
    superStruct['videoPixelSize'] = results.attrs['videoPixelSize']
  if 'pathToOriginalVideo' in results.attrs:
    superStruct['pathToOriginalVideo'] = results.attrs['pathToOriginalVideo']
  superStruct['wellPoissMouv'] = []
  superStruct['wellPositions'] = [dict(results[f'wellPositions/well{idx}'].attrs) for idx in range(len(results['wellPositions']))]
  for wellIdx in range(len(superStruct['wellPositions'])):
    wellData = []
    superStruct['wellPoissMouv'].append(wellData)
    wellGroup = results[f'dataForWell{wellIdx}']
    for animalIdx in range(len(wellGroup)):
      animalGroup = wellGroup[f'dataForAnimal{animalIdx}']
      animalData = []
      wellData.append(animalData)
      if 'dataPerFrame' not in animalGroup:
        continue
      dataGroup = animalGroup['dataPerFrame']
      HeadPos = dataGroup['HeadPos'][:]
      TailPosX = dataGroup['TailPosX'][:]
      TailPosY = dataGroup['TailPosY'][:]
      TailX_VideoReferential = np.column_stack([HeadPos['X']] + [TailPosX[col] for col in dataGroup['TailPosX'].attrs['columns']])
      TailY_VideoReferential = np.column_stack([HeadPos['Y']] + [TailPosY[col] for col in dataGroup['TailPosY'].attrs['columns']])
      TailAngle = dataGroup['TailAngle'][:]
      Heading = dataGroup['Heading'][:]
      curvature = None if 'curvature' not in dataGroup else dataGroup['curvature'][:]
      for boutIdx in range(animalGroup['listOfBouts'].attrs['numberOfBouts']):
        boutGroup = animalGroup[f'listOfBouts/bout{boutIdx}']
        boutData = {'AnimalNumber': animalIdx, 'BoutStart': boutGroup.attrs['BoutStart'], 'BoutEnd': boutGroup.attrs['BoutEnd']}
        if boutGroup.attrs.get('flag'):
          boutData['flag'] = 1
        animalData.append(boutData)
        for data in boutGroup:
          if data == 'additionalKinematicParametersPerBout':
            continue
          boutData[data] = boutGroup[data][:].tolist()
        start = boutGroup.attrs['BoutStart'] - superStruct["firstFrame"]
        end = boutGroup.attrs['BoutEnd'] - superStruct["firstFrame"] + 1
        boutData['TailX_VideoReferential'] = TailX_VideoReferential[start:end].tolist()
        boutData['TailY_VideoReferential'] = TailY_VideoReferential[start:end].tolist()
        boutData['TailAngle_Raw'] = TailAngle[start:end].tolist()
        boutData['Heading'] = Heading[start:end].tolist()
        boutData['HeadX'] = HeadPos['X'][start:end].tolist()
        boutData['HeadY'] = HeadPos['Y'][start:end].tolist()
        if curvature is not None:
          boutData['curvature'] = curvature[start:end].tolist()
  return superStruct



import math

import cv2

import zebrazoom.code.tracking
import zebrazoom.code.tracking.customTrackingImplementations
import zebrazoom.code.util as util
from zebrazoom.code.tracking.customTrackingImplementations.fastFishTracking.utilities import calculateAngle
from zebrazoom.code.adjustHyperparameters import adjustHyperparameters

from PyQt5.QtCore import Qt, QAbstractListModel, QModelIndex


baseClass = zebrazoom.code.tracking.get_tracking_method('fastFishTracking.tracking')


class _CustomListModel(QAbstractListModel):
  def __init__(self, list_, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.itemlist = list_
    self._baseItemCount = len(list_)

  def rowCount(self, parent=QModelIndex()):
    itemCount = len(self.itemlist)
    return itemCount * 2 - 1

  def data(self, index, role):
    if not index.isValid():
      return None
    if role == Qt.AccessibleDescriptionRole and index.row() % 2:
      return 'separator'
    if role != Qt.DisplayRole:
      return None
    idx = index.row() // 2
    currentSteps = len(self.itemlist) - self._baseItemCount
    stepIdx = idx - self.itemlist.index('steps')
    if 0 < stepIdx <= currentSteps:
      return f'Segment length between previous and next point: option {stepIdx}'
    name = self.itemlist[idx]
    return util.PRETTY_PARAMETER_NAMES.get(name, name)

  def updateSteps(self, newSteps):
    listSize = len(self.itemlist)
    currentSteps = listSize - self._baseItemCount
    lastStepIdx = self.itemlist.index('steps') + currentSteps
    difference = newSteps - currentSteps
    if difference > 0:
      startIdx = lastStepIdx * 2 - 1
      self.beginInsertRows(QModelIndex(), startIdx, startIdx + difference * 2 - 1)
      self.itemlist[lastStepIdx+1:lastStepIdx+1] = (f'Step {currentSteps + 1 + idx}' for idx in range(difference))
      self.endInsertRows()
    elif difference < 0:
      startIdx = (lastStepIdx + difference + 2) - 1
      self.beginRemoveRows(QModelIndex(), startIdx, startIdx - difference * 2 - 1)
      del self.itemlist[startIdx:startIdx-difference]
      self.endRemoveRows()


class GUITracking(baseClass):
  def _adjustParameters(self, i, frame, unprocessedFrame, widgets):
    if not self._hyperparameters['adjustFreelySwimTracking']:
      return None
    assert self._hyperparameters['onlyTrackThisOneWell'] != -1

    def trackingAlgorithmChanged(idx, hyperparameters):
      params = ('dualDirectionTailDetection', 'dualDirectionRemoveShortestDirectionFromHead')
      paramsToSet = {'dualDirectionTailDetection'} if idx == 1 else {'dualDirectionTailDetection', 'dualDirectionRemoveShortestDirectionFromHead'} if idx == 2 else set()
      for param in params:
        if param in paramsToSet:
          hyperparameters[param] = 1
        else:
          if param in hyperparameters:
            del hyperparameters[param]

    def getTrackingAlgorithmIndex(hyperparameters):
      params = (None, 'dualDirectionTailDetection', 'dualDirectionRemoveShortestDirectionFromHead')
      for param in reversed(params[1:]):
        if hyperparameters.get(param, False):
          return params.index(param)
      return 0

    hyperparametersListNames = ["maxDepth", "paramGaussianBlur", "headEmbededParamTailDescentPixThreshStop", _CustomListModel(['steps', "thetaDiffAccept", "thetaDiffAcceptAfterAuthorizedRelativeLengthTailEnd", "thetaDiffAcceptAfterAuthorizedRelativeLengthTailEnd2", 'nbList', 'nbListAfterAuthorizedRelativeLengthTailEnd', 'nbListAfterAuthorizedRelativeLengthTailEnd2', 'authorizedRelativeLengthTailEnd', 'authorizedRelativeLengthTailEnd2', 'maximumMedianValueOfAllPointsAlongTheTail', 'minimumHeadPixelValue', 'nbTailPoints', 'paramGaussianBlurForHeadPosition']), ('dualDirectionTailDetection', 'dualDirectionRemoveShortestDirectionFromHead')]
    organizationTab = [
    [1, 50, "Target length of the tail"],
    [1, 30, "Window of gaussian blur filter applied on the image"],
    [1, 300, "Maximum pixel intensity authorized for a pixel to be considered inside the tail"],
    ([1, 10, 'List of possible number of pixels between two subsequent points along the tail'],
     [0.01, 6, 'Maximum authorized angle difference between two subsequent segments in the first portion of the tail (in radians)'],
     [0.01, 6, 'Maximum authorized angle difference between two subsequent segments in the second portion of the tail  (in radians)'],
     [0.01, 6, 'Maximum authorized angle difference between two subsequent segments in the third portion of the tail  (in radians)'],
     [1, 30, 'Number of "candidates" points considered for next point along the tail in the first portion of the tail'],
     [1, 30, 'Number of "candidates" points considered for next point along the tail in the second portion of the tail'],
     [1, 30, 'Number of "candidates" points considered for next point along the tail in the third portion of the tail'],
     [0, 1, 'Cut off relative location between first and second tail segment (between 0 and 1)'],
     [0, 1, 'Cut off relative location between second and third tail segment (between 0 and 1)'],
     [1, 300, 'Maximum median pixel value of all points along the tail in order for the tail tracking to be accepted'],
     [1, 300, 'Maximum pixel value authorized for a point to be considered as the head of the animal'],
     [1, 20, 'Number of points along the tail in the output data'],
     [1, 30, "Window of gaussian blur filter applied on the image when calculating head position"],),
    ['Tracking algorithm', ('Default algorithm: tail descent by segments from head', 'Descent by segments from head in both directions', 'Descent by segments from head in both directions: remove the shortest direction'), trackingAlgorithmChanged, getTrackingAlgorithmIndex],]

    title = "Adjust parameters in order for the background to be white and the animals to be gray/black."

    frame2 = unprocessedFrame if widgets is not None and 'unprocessedFrameCheckbox' in widgets and widgets['unprocessedFrameCheckbox'].isChecked() else cv2.cvtColor(frame, cv2.COLOR_GRAY2RGB)
    zoomInCheckbox = None if widgets is None or 'zoomInCheckbox' not in widgets else widgets['zoomInCheckbox']
    zoomInCoordinates = None
    wellNumber = self._hyperparameters['onlyTrackThisOneWell']
    nbTailPoints = self._hyperparameters['nbTailPoints']
    for k in range(0, self._hyperparameters["nbAnimalsPerWell"]):
      output = self._trackingDataPerWell[wellNumber]
      if self._hyperparameters["trackTail"] == 1:
        for j in range(0, nbTailPoints):
          x = int(output[k, i-self._firstFrame][j][0])
          y = int(output[k, i-self._firstFrame][j][1])
          cv2.circle(frame2, (x, y), 1, (0, 255, 0),   -1)
      x = int(output[k, i-self._firstFrame][nbTailPoints-1][0])
      y = int(output[k, i-self._firstFrame][nbTailPoints-1][1])
      cv2.circle(frame2, (x, y), 2, (0, 0, 255),   -1)
      x = output[k, i-self._firstFrame][0][0]
      y = output[k, i-self._firstFrame][0][1]
      if zoomInCheckbox is not None and zoomInCheckbox.isChecked() and zoomInCheckbox.getAnimalIdx() == k:
        zoomInCoordinates = (x, y)

    if zoomInCoordinates is not None:
      x, y = zoomInCoordinates
      halfLength = 125.5
      xmin = int(x - halfLength)
      xmax = int(x + halfLength)
      ymin = int(y - halfLength)
      ymax = int(y + halfLength)

      x = max(xmin, 0)
      y = max(ymin, 0)
      lengthX = xmax - xmin
      lengthY = ymax - ymin

      frameHeight, frameWidth = frame.shape[:2]
      if y + lengthY >= frameHeight:
        lengthY = frameHeight - y - 1
      if x + lengthX >= frameWidth:
        lengthX = frameWidth - x - 1

      frame2 = frame2[y:y+lengthY, x:x+lengthX]

    return adjustHyperparameters(i, self._hyperparameters, hyperparametersListNames, frame2, title, organizationTab, widgets, addZoomCheckbox=True, addUnprocessedFrameCheckbox=True)


zebrazoom.code.tracking.register_tracking_method('fastFishTracking.tracking', GUITracking)


import struct
import csv
import mmap
import datetime
import array
import logging
import os
import csv
import glob
import time
import configparser
import numpy as np
import cv2
from pathlib import Path
import platform

def sqb_convert_to_avi(path, videoName, codec='HFYU', lastFrame=-1):
  """
      Lecture du fichier binaire de séquence sqb
      Les données sont représentées par la structure en C suivante :
          typedef struct
          {   
              long offset;        // 4 bits -> + 4 bits vides car mémoire alignée
              double TimeStamp;   // 8 bits
              int binfile;        // 4 bits -> + 4 bits vides car mémoire alignée
          } IMGDATA;
  """
  
  # Checking that path and video exists
  if not(os.path.exists(os.path.join(path, videoName + '.seq'))):
    print("Path or video name is incorrect for", os.path.join(path, videoName + '.seq'))
    return 0
  
  seq_path  = os.path.join(path, videoName + '.seq')
  path2 = Path(path)
  avi_path  = os.path.join(path2.parent, videoName + '.avi')

  cfg = configparser.ConfigParser()
  cfg.read(seq_path)
  
  width = int(cfg.get('Sequence Settings', 'Width'))
  height = int(cfg.get('Sequence Settings', 'Height'))
  bpp = int(cfg.get('Sequence Settings', 'BytesPerPixel'))
  num_images = cfg.get('Sequence Settings', 'Number of files')
  bin_file = cfg.get('Sequence Settings', 'Bin File')
  sqb_path = seq_path.replace('.seq', '.sqb')
  
  out = cv2.VideoWriter(avi_path, cv2.VideoWriter_fourcc(codec[0],codec[1],codec[2],codec[3]), 10, (width, height))
  
  pathstr = os.path.dirname(seq_path)
  if lastFrame <= 0:
    lastFrame = int(num_images)
  
  with open(sqb_path,'rb') as f :
    for i in range(0, lastFrame):
    
      if (i % 500 == 0):
        print("image " + str(i) + " out of " + str(lastFrame) + " in total")
    
      if int(platform.system() == "Linux"):
        offset = struct.unpack('l', f.read(8))
        timestamp = struct.unpack('d', f.read(8))
        binfile = struct.unpack('i', f.read(4))
        padding = f.read(4)
      else:
        offset = struct.unpack('l', f.read(4))
        padding = f.read(4)
        timestamp = struct.unpack('d', f.read(8))
        binfile = struct.unpack('i', f.read(4))
        padding = f.read(4)
      
      bin_path = os.path.join("%s" % (pathstr), "%s%0.5d.bin" % (bin_file, binfile[0]))
      
      f_bin = open(bin_path, 'rb')
      f_bin.seek(offset[0], os.SEEK_SET)
      
      bytes = f_bin.read(height*width*bpp)
      
      if bpp == 2:
        buffer = np.frombuffer(bytes, dtype=np.uint16)
      else:
        buffer = np.frombuffer(bytes, dtype=np.uint8)
      
      if len(buffer) == 3*height*width:
        nparr2 = buffer.reshape(height, width, 3)
      else:
        nparr2 = buffer.reshape(height, width)
        nparr2 = cv2.cvtColor(nparr2, cv2.COLOR_GRAY2RGB)
        
      out.write(nparr2)
      
      f_bin.close()
      
  out.release()



import os
import sys

import zebrazoom.code.paths as paths


def launchZebraZoom(args):
  from zebrazoom.GUIAllPy import ZebraZoomApp
  print("The data produced by ZebraZoom can be found in the folder: " + paths.getDefaultZZoutputFolder())
  app = ZebraZoomApp(sys.argv)
  sys.exit(app.exec())
  
def selectZZoutput(args):
  from zebrazoom.GUIAllPy import ZebraZoomApp
  app = ZebraZoomApp(sys.argv)
  print("The data produced by ZebraZoom can be found in the folder: " + paths.getDefaultZZoutputFolder())
  app.askForZZoutputLocation()
  sys.exit(app.exec())

def getTailExtremityFirstFrame(args):
  from zebrazoom.getTailExtremityFirstFrame import getTailExtremityFirstFrame
  __spec__ = "ModuleSpec(name='builtins', loader=<class '_frozen_importlib.BuiltinImporter'>)"
  getTailExtremityFirstFrame(args.pathToVideo, args.videoName, args.videoExt, args.configFile, args.hyperparameters)

def recreateSuperStruct(args):
  from zebrazoom.recreateSuperStruct import recreateSuperStruct
  __spec__ = "ModuleSpec(name='builtins', loader=<class '_frozen_importlib.BuiltinImporter'>)"
  recreateSuperStruct(args.pathToVideo, args.videoName, args.videoExt, args.configFile, args.hyperparameters)

def convertSeqToAvi(args):
  from zebrazoom.videoFormatConversion.seq_to_avi import sqb_convert_to_avi
  sqb_convert_to_avi(args.path, args.videoName, args.codec, args.lastFrame)

def convertSeqToAviThenLaunchTracking(args):
  from zebrazoom.videoFormatConversion.seq_to_avi import sqb_convert_to_avi
  from zebrazoom.zebraZoomVideoAnalysis import ZebraZoomVideoAnalysis
  from pathlib import Path
  import time
  path2      = Path(args.path).parent
  print("Launching the convertion from seq to avi")
  sqb_convert_to_avi(args.path, args.videoName, args.codec, args.lastFrame)
  print("small break start")
  time.sleep(2)
  print("Launching the tracking, the data produced by ZebraZoom can be found in the folder: " + paths.getDefaultZZoutputFolder())
  __spec__ = "ModuleSpec(name='builtins', loader=<class '_frozen_importlib.BuiltinImporter'>)"
  ZebraZoomVideoAnalysis(path2, args.videoName, 'avi', args.configFile, args.hyperparameters, useGUI=False).run()

def DL_createMask(args):
  from zebrazoom.code.deepLearningFunctions.labellingFunctions import createMask
  pathToImgFolder = args.pathToImgFolder
  if not(os.path.exists(pathToImgFolder)):
    pathToImgFolder = os.path.join(paths.getDefaultZZoutputFolder(), args.pathToImgFolder, 'PNGImages')
  createMask(pathToImgFolder)

def sleepVsMoving(args):
  from zebrazoom.code.dataPostProcessing.findSleepVsMoving import calculateSleepVsMovingPeriods
  calculateSleepVsMovingPeriods(paths.getDefaultZZoutputFolder(), args.videoName, args.speedThresholdForMoving, args.notMovingNumberOfFramesThresholdForSleep,
                                args.maxDistBetweenTwoPointsInsideSleepingPeriod, args.specifiedStartTime, args.distanceTravelledRollingMedianFilter,
                                args.videoPixelSize, args.videoFPS)

def firstSleepingTimeAfterSpecifiedTime(args):
  from zebrazoom.code.dataPostProcessing.findSleepVsMoving import firstSleepingTimeAfterSpecifiedTime
  pathToZZoutput = paths.getDefaultZZoutputFolder()
  firstSleepingTimeAfterSpecifiedTime(args.pathToZZoutput, args.videoName, args.specifiedTime, args.wellNumber)

def numberOfSleepingAndMovingTimesInTimeRange(args):
  from zebrazoom.code.dataPostProcessing.findSleepVsMoving import numberOfSleepingAndMovingTimesInTimeRange
  numberOfSleepingAndMovingTimesInTimeRange(paths.getDefaultZZoutputFolder(), args.videoName, args.specifiedStartTime, args.specifiedEndTime, args.wellNumber)

def numberOfSleepBoutsInTimeRange(args):
  from zebrazoom.code.dataPostProcessing.findSleepVsMoving import numberOfSleepBoutsInTimeRange
  numberOfSleepBoutsInTimeRange(paths.getDefaultZZoutputFolder(), args.videoName, args.minSleepLenghtDurationThreshold, args.wellNumber,
                                args.specifiedStartTime, args.specifiedEndTime)

def calculateNumberOfSfsVsTurnsBasedOnMaxAmplitudeThreshod(args):
  from zebrazoom.dataAnalysis.postProcessingFromCommandLine.postProcessingFromCommandLine import calculateNumberOfSfsVsTurnsBasedOnMaxAmplitudeThreshold
  calculateNumberOfSfsVsTurnsBasedOnMaxAmplitudeThreshold(paths.getRootDataFolder(), args.experimentName, args.thresholdInDegrees)

def kinematicParametersAnalysis(args):
  from zebrazoom.kinematicParametersAnalysis import kinematicParametersAnalysis
  kinematicParametersAnalysis(sys)

def kinematicParametersAnalysisWithMedianPerGenotype(args):
  from zebrazoom.kinematicParametersAnalysis import kinematicParametersAnalysis
  kinematicParametersAnalysis(sys, 1)

def clusteringAnalysis(args):
  from zebrazoom.clusteringAnalysis import clusteringAnalysis
  clusteringAnalysis(sys)

def clusteringAnalysisPerFrame(args):
  from zebrazoom.clusteringAnalysisPerFrame import clusteringAnalysisPerFrame
  clusteringAnalysisPerFrame(sys)

def kinematicParametersAnalysisCenterOfMassOnlyNoBoutsDetection(args):
  from zebrazoom.kinematicParametersAnalysisCenterOfMassOnlyNoBoutsDetection import kinematicParametersAnalysisCenterOfMassOnlyNoBoutsDetection
  kinematicParametersAnalysisCenterOfMassOnlyNoBoutsDetection(sys)

def visualizeMovingAndSleepingTime(args):
  from zebrazoom.code.GUI.readValidationVideo import readValidationVideo
  import pandas as pd
  df = pd.read_excel(os.path.join(paths.getDefaultZZoutputFolder(), args.videoName, "sleepVsMoving_" + args.videoName + ".xlsx"))
  nbWells = int(len(df.columns)/3)

  if args.movingOrSleeping == "movingTime":
    framesToShow = df[["moving_" + str(i) for i in range(0, nbWells)]].to_numpy()
  else:
    assert args.movingOrSleeping == "sleepingTime"
    framesToShow = df[["sleep_" + str(i) for i in range(0, nbWells)]].to_numpy()
  readValidationVideo("", args.videoName, -1, -1, 0, 1, framesToShow)

def createDistanceBetweenFramesExcelFile(args):
  from zebrazoom.dataAnalysis.createCustomDataStructure.createDistanceBetweenFramesExcelFile import createDistanceBetweenFramesExcelFile
  from zebrazoom.GUIAllPy import PlainApplication
  app = PlainApplication(sys.argv)
  createDistanceBetweenFramesExcelFile(paths.getDefaultZZoutputFolder(), args.videoFPS, args.videoPixelSize) # fps, pixelSize

def createDistanceSpeedHeadingDeltaHeadingExcelFile(args):
  from zebrazoom.dataAnalysis.createCustomDataStructure.createDistanceSpeedHeadingDeltaHeadingExcelFile import createDistanceSpeedHeadingDeltaHeadingExcelFile
  from zebrazoom.GUIAllPy import PlainApplication
  app = PlainApplication(sys.argv)
  createDistanceSpeedHeadingDeltaHeadingExcelFile(paths.getDefaultZZoutputFolder(), args.videoFPS, args.videoPixelSize) # fps, pixelSize

def removeLargeInstantaneousDistanceData(args):
  from zebrazoom.dataAnalysis.createCustomDataStructure.removeLargeInstantaneousDistanceData import removeLargeInstantaneousDistanceData
  from zebrazoom.GUIAllPy import PlainApplication
  app = PlainApplication(sys.argv)
  removeLargeInstantaneousDistanceData(paths.getDefaultZZoutputFolder(), args.maxDistance)

def filterLatencyAndMergeBoutsInSameTrials(args):
  from zebrazoom.dataAnalysis.createCustomDataStructure.filterLatencyAndMergeBoutsInSameTrials import filterLatencyAndMergeBoutsInSameTrials
  filterLatencyAndMergeBoutsInSameTrials(args.nameOfExperiment, args.minFrameNumberBoutStart, args.maxFrameNumberBoutStart, args.calculationMethod,
                                         paths.getDefaultZZoutputFolder(), args.dropDuplicates)

def launchActiveLearning(args):
  from zebrazoom.otherScripts.launchActiveLearning import launchActiveLearning
  launchActiveLearning()

def launchOptimalClusterNumberSearch(args):
  from zebrazoom.otherScripts.launchOptimalClusterNumberSearch import launchOptimalClusterNumberSearch
  launchOptimalClusterNumberSearch()

def launchReapplyClustering(args):
  from zebrazoom.otherScripts.launchReapplyClustering import launchReapplyClustering
  launchReapplyClustering()

def createSmallValidationVideosForFlagged(args):
  from zebrazoom.code.createValidationVideo import createSmallValidationVideosForFlagged
  createSmallValidationVideosForFlagged(args.videoName.rstrip(r'\/'), args.offset)

def alternativeKinematicParameterCalculation(args):
  from zebrazoom.otherScripts.alternativeKinematicParameterCalculation import alternativeKinematicParameterCalculation
  alternativeKinematicParameterCalculation(args)

def exit(args):
  from PyQt5.QtCore import QTimer
  from zebrazoom.GUIAllPy import ZebraZoomApp
  app = ZebraZoomApp(sys.argv)
  QTimer.singleShot(0, app.window.close)
  sys.exit(app.exec())

def runVideoAnalysis(args):
  print("The data produced by ZebraZoom can be found in the folder: " + paths.getDefaultZZoutputFolder())
  from zebrazoom.zebraZoomVideoAnalysis import ZebraZoomVideoAnalysis

  if args.useGUI:
    try:
      from zebrazoom.GUIAllPy import PlainApplication
      app = PlainApplication(sys.argv)
    except ImportError:
      args.useGUI = False
      print("GUI not available")
  __spec__ = "ModuleSpec(name='builtins', loader=<class '_frozen_importlib.BuiltinImporter'>)"
  ZebraZoomVideoAnalysis(args.pathToVideo, args.videoName, args.videoExt, args.configFile, args.hyperparameters, useGUI=args.useGUI).run()
